import keras

from keras_hub.src.models.mobilenet.util import adjust_channels
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import ConvNormAct
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import DropPath
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import LayerScale2d
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import RmsNorm2d
from keras_hub.src.models.mobilenetv5.mobilenetv5_utils import num_groups


class UniversalInvertedResidual(keras.layers.Layer):
    """Universal Inverted Residual block.

    This block is a flexible and universal version of the inverted residual
    block, which can be configured to behave like different variants of mobile
    convolutional blocks.

    Args:
        filters: int. The number of output channels.
        dw_kernel_size_start: int. The kernel size for the initial depthwise
            convolution. If 0, this layer is skipped.
        dw_kernel_size_mid: int. The kernel size for the middle depthwise
            convolution. If 0, this layer is skipped.
        dw_kernel_size_end: int. The kernel size for the final depthwise
            convolution. If 0, this layer is skipped.
        stride: int. The stride for the block.
        dilation: int. The dilation rate for convolutions.
        pad_type: str. The padding type for convolutions.
        noskip: bool. If `True`, the skip connection is disabled.
        exp_ratio: float. The expansion ratio for the middle channels.
        act_layer: str. The activation function to use.
        norm_layer: str. The normalization layer to use.
        se_layer: keras.layers.Layer. The Squeeze-and-Excitation layer to use.
        drop_path_rate: float. The stochastic depth rate.
        layer_scale_init_value: float. The initial value for layer scale. If
            `None`, layer scale is not used.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        filters,
        dw_kernel_size_start=0,
        dw_kernel_size_mid=3,
        dw_kernel_size_end=0,
        stride=1,
        dilation=1,
        pad_type="same",
        noskip=False,
        exp_ratio=1.0,
        act_layer="relu",
        norm_layer="batch_norm",
        se_layer=None,
        drop_path_rate=0.0,
        layer_scale_init_value=1e-5,
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.dw_kernel_size_start = dw_kernel_size_start
        self.dw_kernel_size_mid = dw_kernel_size_mid
        self.dw_kernel_size_end = dw_kernel_size_end
        self.stride = stride
        self.dilation = dilation
        self.pad_type = pad_type
        self.noskip = noskip
        self.exp_ratio = exp_ratio
        self.act_layer = act_layer
        self.norm_layer = norm_layer
        self.se_layer = se_layer
        self.drop_path_rate = drop_path_rate
        self.layer_scale_init_value = layer_scale_init_value
        self.data_format = data_format
        self.channel_axis = channel_axis

    def build(self, input_shape):
        super().build(input_shape)
        in_chs = input_shape[self.channel_axis]
        self.has_skip = (
            in_chs == self.filters and self.stride == 1
        ) and not self.noskip
        use_bias = self.norm_layer == "rms_norm"

        if self.dw_kernel_size_start:
            self.dw_start = ConvNormAct(
                in_chs,
                self.dw_kernel_size_start,
                stride=self.stride if not self.dw_kernel_size_mid else 1,
                dilation=self.dilation,
                groups=in_chs,
                pad_type=self.pad_type,
                apply_act=False,
                act_layer=self.act_layer,
                norm_layer=self.norm_layer,
                bias=use_bias,
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
        else:
            self.dw_start = lambda x, training=False: x

        mid_chs = adjust_channels(in_chs * self.exp_ratio)
        self.pw_exp = ConvNormAct(
            mid_chs,
            1,
            pad_type=self.pad_type,
            act_layer=self.act_layer,
            norm_layer=self.norm_layer,
            bias=use_bias,
            data_format=self.data_format,
            channel_axis=self.channel_axis,
            dtype=self.dtype_policy,
        )

        if self.dw_kernel_size_mid:
            self.dw_mid = ConvNormAct(
                mid_chs,
                self.dw_kernel_size_mid,
                stride=self.stride,
                dilation=self.dilation,
                groups=mid_chs,
                pad_type=self.pad_type,
                act_layer=self.act_layer,
                norm_layer=self.norm_layer,
                bias=use_bias,
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
        else:
            self.dw_mid = lambda x, training=False: x
        self.se = (
            self.se_layer(
                filters=mid_chs,
                bottleneck_filters=adjust_channels(mid_chs * 0.25),
                squeeze_activation=self.act_layer,
                excite_activation="sigmoid",
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
            if self.se_layer
            else (lambda x, training=False: x)
        )
        self.pw_proj = ConvNormAct(
            self.filters,
            1,
            pad_type=self.pad_type,
            apply_act=False,
            act_layer=self.act_layer,
            norm_layer=self.norm_layer,
            bias=use_bias,
            data_format=self.data_format,
            channel_axis=self.channel_axis,
            dtype=self.dtype_policy,
        )

        if self.dw_kernel_size_end:
            self.dw_end = ConvNormAct(
                self.filters,
                self.dw_kernel_size_end,
                stride=self.stride
                if not self.dw_kernel_size_start and not self.dw_kernel_size_mid
                else 1,
                dilation=self.dilation,
                groups=self.filters,
                pad_type=self.pad_type,
                apply_act=False,
                act_layer=self.act_layer,
                norm_layer=self.norm_layer,
                bias=use_bias,
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
        else:
            self.dw_end = lambda x, training=False: x

        self.layer_scale = (
            LayerScale2d(
                self.filters,
                self.layer_scale_init_value,
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
            if self.layer_scale_init_value is not None
            else lambda x: x
        )
        self.drop_path = (
            DropPath(self.drop_path_rate, dtype=self.dtype_policy)
            if self.drop_path_rate > 0.0
            else (lambda x, training=False: x)
        )
        current_shape = input_shape
        if hasattr(self.dw_start, "build"):
            self.dw_start.build(current_shape)
            current_shape = self.dw_start.compute_output_shape(current_shape)
        self.pw_exp.build(current_shape)
        current_shape = self.pw_exp.compute_output_shape(current_shape)
        if hasattr(self.dw_mid, "build"):
            self.dw_mid.build(current_shape)
            current_shape = self.dw_mid.compute_output_shape(current_shape)
        if hasattr(self.se, "build"):
            self.se.build(current_shape)
        self.pw_proj.build(current_shape)
        current_shape = self.pw_proj.compute_output_shape(current_shape)
        if hasattr(self.dw_end, "build"):
            self.dw_end.build(current_shape)
            current_shape = self.dw_end.compute_output_shape(current_shape)
        if hasattr(self.layer_scale, "build"):
            self.layer_scale.build(current_shape)

    def call(self, x, training=False):
        shortcut = x
        x = self.dw_start(x, training=training)
        x = self.pw_exp(x, training=training)
        x = self.dw_mid(x, training=training)
        x = self.se(x, training=training)
        x = self.pw_proj(x, training=training)
        x = self.dw_end(x, training=training)
        x = self.layer_scale(x)
        if self.has_skip:
            x = self.drop_path(x, training=training) + shortcut
        return x

    def compute_output_shape(self, input_shape):
        current_shape = input_shape
        if hasattr(self.dw_start, "compute_output_shape"):
            current_shape = self.dw_start.compute_output_shape(current_shape)
        current_shape = self.pw_exp.compute_output_shape(current_shape)
        if hasattr(self.dw_mid, "compute_output_shape"):
            current_shape = self.dw_mid.compute_output_shape(current_shape)
        current_shape = self.pw_proj.compute_output_shape(current_shape)
        if hasattr(self.dw_end, "compute_output_shape"):
            current_shape = self.dw_end.compute_output_shape(current_shape)
        return current_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "dw_kernel_size_start": self.dw_kernel_size_start,
                "dw_kernel_size_mid": self.dw_kernel_size_mid,
                "dw_kernel_size_end": self.dw_kernel_size_end,
                "stride": self.stride,
                "dilation": self.dilation,
                "pad_type": self.pad_type,
                "noskip": self.noskip,
                "exp_ratio": self.exp_ratio,
                "act_layer": self.act_layer,
                "norm_layer": self.norm_layer,
                "se_layer": keras.saving.serialize_keras_object(self.se_layer),
                "drop_path_rate": self.drop_path_rate,
                "layer_scale_init_value": self.layer_scale_init_value,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config["se_layer"] = keras.saving.deserialize_keras_object(
            config.pop("se_layer")
        )
        return cls(**config)


class EdgeResidual(keras.layers.Layer):
    """Edge Residual block.

    This block is designed for efficiency on edge devices. It is a variant of
    the inverted residual block that uses a single expansion convolution.

    Args:
        filters: int. The number of output channels.
        exp_kernel_size: int. The kernel size for the expansion convolution.
        stride: int. The stride for the block.
        dilation: int. The dilation rate for convolutions.
        group_size: int. The group size for grouped convolutions.
        pad_type: str. The padding type for convolutions.
        expansion_in_chs: int. If greater than 0, forces the number of input
            channels for the expansion.
        noskip: bool. If `True`, the skip connection is disabled.
        exp_ratio: float. The expansion ratio for the middle channels.
        pw_kernel_size: int. The kernel size for the pointwise convolution.
        act_layer: str. The activation function to use.
        norm_layer: str. The normalization layer to use.
        se_layer: keras.layers.Layer. The Squeeze-and-Excitation layer to use.
        drop_path_rate: float. The stochastic depth rate.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        filters,
        exp_kernel_size=3,
        stride=1,
        dilation=1,
        group_size=0,
        pad_type="same",
        expansion_in_chs=0,
        noskip=False,
        exp_ratio=1.0,
        pw_kernel_size=1,
        act_layer="relu",
        norm_layer="batch_norm",
        se_layer=None,
        drop_path_rate=0.0,
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.exp_kernel_size = exp_kernel_size
        self.stride = stride
        self.dilation = dilation
        self.group_size = group_size
        self.pad_type = pad_type
        self.expansion_in_chs = expansion_in_chs
        self.noskip = noskip
        self.exp_ratio = exp_ratio
        self.pw_kernel_size = pw_kernel_size
        self.act_layer = act_layer
        self.norm_layer = norm_layer
        self.se_layer = se_layer
        self.drop_path_rate = drop_path_rate
        self.data_format = data_format
        self.channel_axis = channel_axis

    def build(self, input_shape):
        super().build(input_shape)
        in_chs = input_shape[self.channel_axis]
        self.has_skip = (
            in_chs == self.filters and self.stride == 1
        ) and not self.noskip
        if self.expansion_in_chs > 0:
            mid_chs = adjust_channels(self.expansion_in_chs * self.exp_ratio)
        else:
            mid_chs = adjust_channels(in_chs * self.exp_ratio)
        groups = num_groups(self.group_size, mid_chs)
        use_bias = self.norm_layer == "rms_norm"
        self.conv_exp = ConvNormAct(
            mid_chs,
            self.exp_kernel_size,
            stride=self.stride,
            dilation=self.dilation,
            groups=groups,
            pad_type=self.pad_type,
            norm_layer=self.norm_layer,
            act_layer=self.act_layer,
            bias=use_bias,
            data_format=self.data_format,
            channel_axis=self.channel_axis,
            dtype=self.dtype_policy,
        )
        self.se = (
            self.se_layer(
                filters=mid_chs,
                bottleneck_filters=adjust_channels(mid_chs * 0.25),
                squeeze_activation=self.act_layer,
                excite_activation="sigmoid",
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
            if self.se_layer
            else (lambda x, training=False: x)
        )
        self.conv_pwl = ConvNormAct(
            self.filters,
            self.pw_kernel_size,
            pad_type=self.pad_type,
            apply_act=False,
            norm_layer=self.norm_layer,
            act_layer=self.act_layer,
            bias=use_bias,
            data_format=self.data_format,
            channel_axis=self.channel_axis,
            dtype=self.dtype_policy,
        )
        self.drop_path = (
            DropPath(self.drop_path_rate, dtype=self.dtype_policy)
            if self.drop_path_rate > 0.0
            else (lambda x, training=False: x)
        )
        self.conv_exp.build(input_shape)
        conv_exp_output_shape = self.conv_exp.compute_output_shape(input_shape)
        if hasattr(self.se, "build"):
            self.se.build(conv_exp_output_shape)
        self.conv_pwl.build(conv_exp_output_shape)

    def call(self, x, training=False):
        shortcut = x
        x = self.conv_exp(x, training=training)
        x = self.se(x, training=training)
        x = self.conv_pwl(x, training=training)
        if self.has_skip:
            x = self.drop_path(x, training=training) + shortcut
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "exp_kernel_size": self.exp_kernel_size,
                "stride": self.stride,
                "dilation": self.dilation,
                "group_size": self.group_size,
                "pad_type": self.pad_type,
                "expansion_in_chs": self.expansion_in_chs,
                "noskip": self.noskip,
                "exp_ratio": self.exp_ratio,
                "pw_kernel_size": self.pw_kernel_size,
                "act_layer": self.act_layer,
                "norm_layer": self.norm_layer,
                "se_layer": keras.saving.serialize_keras_object(self.se_layer),
                "drop_path_rate": self.drop_path_rate,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config["se_layer"] = keras.saving.deserialize_keras_object(
            config.pop("se_layer")
        )
        return cls(**config)


class CondConvResidual(keras.layers.Layer):
    """Conditionally Parameterized Convolutional Residual block.

    This block uses a routing function to dynamically select and combine
    different convolutional experts based on the input.

    Args:
        filters: int. The number of output channels.
        dw_kernel_size: int. The kernel size for the depthwise convolution.
        stride: int. The stride for the block.
        dilation: int. The dilation rate for convolutions.
        pad_type: str. The padding type for convolutions.
        noskip: bool. If `True`, the skip connection is disabled.
        exp_ratio: float. The expansion ratio for the middle channels.
        exp_kernel_size: int. The kernel size for the expansion convolution.
        pw_kernel_size: int. The kernel size for the pointwise convolution.
        act_layer: str. The activation function to use.
        se_layer: keras.layers.Layer. The Squeeze-and-Excitation layer to use.
        num_experts: int. The number of experts to use.
        drop_path_rate: float. The stochastic depth rate.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        filters,
        dw_kernel_size=3,
        stride=1,
        dilation=1,
        pad_type="same",
        noskip=False,
        exp_ratio=1.0,
        exp_kernel_size=1,
        pw_kernel_size=1,
        act_layer="relu",
        se_layer=None,
        num_experts=0,
        drop_path_rate=0.0,
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.dw_kernel_size = dw_kernel_size
        self.stride = stride
        self.dilation = dilation
        self.pad_type = pad_type
        self.noskip = noskip
        self.exp_ratio = exp_ratio
        self.exp_kernel_size = exp_kernel_size
        self.pw_kernel_size = pw_kernel_size
        self.act_layer = act_layer
        self.se_layer = se_layer
        self.num_experts = num_experts
        self.drop_path_rate = drop_path_rate
        self.data_format = data_format
        self.channel_axis = channel_axis
        self.conv_kernel_initializer = keras.initializers.VarianceScaling(
            scale=2.0, mode="fan_out", distribution="untruncated_normal"
        )
        self.dense_kernel_initializer = keras.initializers.VarianceScaling(
            scale=1.0, mode="fan_in", distribution="uniform"
        )
        self.bias_initializer = "zeros"

    def build(self, input_shape):
        super().build(input_shape)
        in_chs = input_shape[self.channel_axis]
        self.has_skip = (
            in_chs == self.filters and self.stride == 1
        ) and not self.noskip
        mid_chs = adjust_channels(in_chs * self.exp_ratio)
        self.routing_fn = keras.layers.Dense(
            self.num_experts,
            dtype=self.dtype_policy,
            kernel_initializer=self.dense_kernel_initializer,
            bias_initializer=self.bias_initializer,
        )
        self.pool = keras.layers.GlobalAveragePooling2D(
            data_format=self.data_format, dtype=self.dtype_policy
        )
        self.conv_pw_experts = [
            keras.layers.Conv2D(
                filters=mid_chs,
                kernel_size=self.exp_kernel_size,
                padding=self.pad_type,
                use_bias=True,
                data_format=self.data_format,
                name=f"conv_pw_expert_{i}",
                kernel_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
            for i in range(self.num_experts)
        ]
        self.conv_dw_experts = [
            keras.layers.DepthwiseConv2D(
                kernel_size=self.dw_kernel_size,
                strides=self.stride,
                padding=self.pad_type,
                dilation_rate=self.dilation,
                use_bias=True,
                data_format=self.data_format,
                name=f"conv_dw_expert_{i}",
                depthwise_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
            for i in range(self.num_experts)
        ]
        self.conv_pwl_experts = [
            keras.layers.Conv2D(
                filters=self.filters,
                kernel_size=self.pw_kernel_size,
                padding=self.pad_type,
                use_bias=True,
                data_format=self.data_format,
                name=f"conv_pwl_expert_{i}",
                kernel_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
            for i in range(self.num_experts)
        ]
        self.bn1 = keras.layers.BatchNormalization(
            axis=self.channel_axis,
            dtype=self.dtype_policy,
            gamma_initializer="ones",
            beta_initializer="zeros",
        )
        self.act1 = keras.layers.Activation(
            self.act_layer, dtype=self.dtype_policy
        )
        self.bn2 = keras.layers.BatchNormalization(
            axis=self.channel_axis,
            dtype=self.dtype_policy,
            gamma_initializer="ones",
            beta_initializer="zeros",
        )
        self.act2 = keras.layers.Activation(
            self.act_layer, dtype=self.dtype_policy
        )
        self.bn3 = keras.layers.BatchNormalization(
            axis=self.channel_axis,
            dtype=self.dtype_policy,
            gamma_initializer="ones",
            beta_initializer="zeros",
        )
        self.se = (
            self.se_layer(
                filters=mid_chs,
                bottleneck_filters=adjust_channels(mid_chs * 0.25),
                squeeze_activation=self.act_layer,
                excite_activation="sigmoid",
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
            if self.se_layer
            else (lambda x, training=False: x)
        )
        self.drop_path = (
            DropPath(self.drop_path_rate, dtype=self.dtype_policy)
            if self.drop_path_rate > 0.0
            else (lambda x, training=False: x)
        )
        pooled_shape = self.pool.compute_output_shape(input_shape)
        self.routing_fn.build(pooled_shape)
        for expert in self.conv_pw_experts:
            expert.build(input_shape)
        pw_out_shape = self.conv_pw_experts[0].compute_output_shape(input_shape)
        self.bn1.build(pw_out_shape)
        for expert in self.conv_dw_experts:
            expert.build(pw_out_shape)
        dw_out_shape = self.conv_dw_experts[0].compute_output_shape(
            pw_out_shape
        )
        self.bn2.build(dw_out_shape)
        if hasattr(self.se, "build"):
            self.se.build(dw_out_shape)
        for expert in self.conv_pwl_experts:
            expert.build(dw_out_shape)
        pwl_out_shape = self.conv_pwl_experts[0].compute_output_shape(
            dw_out_shape
        )
        self.bn3.build(pwl_out_shape)

    def _apply_cond_conv(self, x, experts, routing_weights):
        outputs = []
        for i, expert in enumerate(experts):
            expert_out = expert(x)
            weight = keras.ops.reshape(routing_weights[:, i], (-1, 1, 1, 1))
            outputs.append(expert_out * weight)
        return keras.ops.sum(outputs, axis=0)

    def call(self, x, training=False):
        shortcut = x
        pooled_inputs = self.pool(x)
        routing_weights = keras.activations.sigmoid(
            self.routing_fn(pooled_inputs)
        )
        x = self._apply_cond_conv(x, self.conv_pw_experts, routing_weights)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        x = self._apply_cond_conv(x, self.conv_dw_experts, routing_weights)
        x = self.bn2(x, training=training)
        x = self.act2(x)
        x = self.se(x, training=training)
        x = self._apply_cond_conv(x, self.conv_pwl_experts, routing_weights)
        x = self.bn3(x, training=training)
        if self.has_skip:
            x = self.drop_path(x, training=training) + shortcut
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "dw_kernel_size": self.dw_kernel_size,
                "stride": self.stride,
                "dilation": self.dilation,
                "pad_type": self.pad_type,
                "noskip": self.noskip,
                "exp_ratio": self.exp_ratio,
                "exp_kernel_size": self.exp_kernel_size,
                "pw_kernel_size": self.pw_kernel_size,
                "act_layer": self.act_layer,
                "se_layer": keras.saving.serialize_keras_object(self.se_layer),
                "num_experts": self.num_experts,
                "drop_path_rate": self.drop_path_rate,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config["se_layer"] = keras.saving.deserialize_keras_object(
            config.pop("se_layer")
        )
        return cls(**config)


class MobileNetV5MultiScaleFusionAdapter(keras.layers.Layer):
    """Multi-Scale Fusion Adapter for MobileNetV5.

    This layer fuses feature maps from different scales of the backbone,
    concatenates them, processes them through a FFN (Feed-Forward Network),
    and then resizes the output to a target resolution.

    Args:
        in_chs: list of int. A list of channel counts for each input feature
            map.
        filters: int. The number of output channels.
        output_resolution: int or tuple. The target output resolution.
        expansion_ratio: float. The expansion ratio for the FFN.
        interpolation_mode: str. The interpolation mode for upsampling feature
            maps.
        layer_scale_init_value: float. The initial value for layer scale. If
            `None`, layer scale is not used.
        noskip: bool. If `True`, the skip connection in the FFN is disabled.
        act_layer: str. The activation function to use.
        norm_layer: str. The normalization layer to use.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        in_chs,
        filters,
        output_resolution,
        expansion_ratio=2.0,
        interpolation_mode="nearest",
        layer_scale_init_value=None,
        noskip=True,
        act_layer="gelu",
        norm_layer="rms_norm",
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.in_chs = in_chs
        self.filters = filters
        self.output_resolution_arg = output_resolution
        self.expansion_ratio = expansion_ratio
        self.interpolation_mode = interpolation_mode
        self.layer_scale_init_value = layer_scale_init_value
        self.noskip = noskip
        self.act_layer = act_layer
        self.norm_layer_name = norm_layer
        self.data_format = data_format
        self.channel_axis = channel_axis
        self.in_channels = sum(in_chs)
        if isinstance(output_resolution, int):
            self.output_resolution = (output_resolution, output_resolution)
        else:
            self.output_resolution = output_resolution
        self.ffn = UniversalInvertedResidual(
            filters=self.filters,
            dw_kernel_size_mid=0,
            exp_ratio=expansion_ratio,
            act_layer=act_layer,
            norm_layer=norm_layer,
            noskip=noskip,
            layer_scale_init_value=layer_scale_init_value,
            data_format=self.data_format,
            channel_axis=self.channel_axis,
            dtype=self.dtype_policy,
        )
        if norm_layer == "rms_norm":
            self.norm = RmsNorm2d(
                self.filters,
                data_format=self.data_format,
                gamma_initializer="ones",
                channel_axis=self.channel_axis,
                dtype=self.dtype_policy,
            )
        else:
            self.norm = keras.layers.BatchNormalization(
                axis=self.channel_axis,
                gamma_initializer="ones",
                beta_initializer="zeros",
                dtype=self.dtype_policy,
            )

    def build(self, input_shape):
        super().build(input_shape)
        ffn_input_shape = list(input_shape[0])
        if self.data_format == "channels_first":
            ffn_input_shape[1] = self.in_channels
        else:
            ffn_input_shape[-1] = self.in_channels
        self.ffn.build(tuple(ffn_input_shape))
        norm_input_shape = self.ffn.compute_output_shape(tuple(ffn_input_shape))
        self.norm.build(norm_input_shape)

    def call(self, inputs, training=False):
        shape_hr = keras.ops.shape(inputs[0])
        if self.data_format == "channels_first":
            high_resolution = (shape_hr[2], shape_hr[3])
        else:
            high_resolution = (shape_hr[1], shape_hr[2])
        resized_inputs = []
        for img in inputs:
            if self.data_format == "channels_first":
                img_transposed = keras.ops.transpose(img, (0, 2, 3, 1))
            else:
                img_transposed = img
            img_resized = keras.ops.image.resize(
                img_transposed,
                size=high_resolution,
                interpolation=self.interpolation_mode,
            )
            if self.data_format == "channels_first":
                resized_inputs.append(
                    keras.ops.transpose(img_resized, (0, 3, 1, 2))
                )
            else:
                resized_inputs.append(img_resized)
        channel_cat_imgs = keras.ops.concatenate(
            resized_inputs, axis=self.channel_axis
        )
        img = self.ffn(channel_cat_imgs, training=training)
        if (
            high_resolution[0] != self.output_resolution[0]
            or high_resolution[1] != self.output_resolution[1]
        ):
            h_in, w_in = high_resolution
            h_out, w_out = self.output_resolution
            if h_in % h_out == 0 and w_in % w_out == 0:
                h_stride = h_in // h_out
                w_stride = w_in // w_out
                img = keras.ops.nn.average_pool(
                    img,
                    pool_size=(h_stride, w_stride),
                    strides=(h_stride, w_stride),
                    padding="valid",
                    data_format=self.data_format,
                )
            else:
                if self.data_format == "channels_first":
                    img_transposed = keras.ops.transpose(img, (0, 2, 3, 1))
                else:
                    img_transposed = img
                img_resized = keras.ops.image.resize(
                    img_transposed,
                    size=self.output_resolution,
                    interpolation="bilinear",
                )
                if self.data_format == "channels_first":
                    img = keras.ops.transpose(img_resized, (0, 3, 1, 2))
                else:
                    img = img_resized
        img = self.norm(img, training=training)
        return img

    def compute_output_shape(self, input_shape):
        batch_size = input_shape[0][0]
        if self.data_format == "channels_first":
            return (
                batch_size,
                self.filters,
                self.output_resolution[0],
                self.output_resolution[1],
            )
        else:
            return (
                batch_size,
                self.output_resolution[0],
                self.output_resolution[1],
                self.filters,
            )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "in_chs": self.in_chs,
                "filters": self.filters,
                "output_resolution": self.output_resolution_arg,
                "expansion_ratio": self.expansion_ratio,
                "interpolation_mode": self.interpolation_mode,
                "layer_scale_init_value": self.layer_scale_init_value,
                "noskip": self.noskip,
                "act_layer": self.act_layer,
                "norm_layer": self.norm_layer_name,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config



import re
from copy import deepcopy

from keras_hub.src.models.mobilenet.mobilenet_backbone import ConvBnActBlock
from keras_hub.src.models.mobilenet.mobilenet_backbone import DepthwiseConvBlock
from keras_hub.src.models.mobilenet.mobilenet_backbone import (
    InvertedResidualBlock,
)
from keras_hub.src.models.mobilenet.util import adjust_channels
from keras_hub.src.models.mobilenetv5.mobilenetv5_attention import (
    MobileAttention,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_blocks import CondConvResidual
from keras_hub.src.models.mobilenetv5.mobilenetv5_blocks import EdgeResidual
from keras_hub.src.models.mobilenetv5.mobilenetv5_blocks import (
    UniversalInvertedResidual,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_utils import parse_ksize
from keras_hub.src.models.mobilenetv5.mobilenetv5_utils import round_channels


def decode_block_str(block_str):
    assert isinstance(block_str, str)
    ops = block_str.split("_")
    block_type = ops[0]
    ops = ops[1:]
    options = {}
    skip = None
    for op in ops:
        if op == "noskip":
            skip = False
        elif op == "skip":
            skip = True
        elif op.startswith("n"):
            key = op[0]
            v = op[1:]
            options[key] = v if v else "relu"
        else:
            splits = re.split(r"(\d.*)", op)
            if len(splits) >= 2:
                key, value = splits[:2]
                options[key] = value

    act_layer = options.get("n", "gelu")
    num_repeat = int(options["r"])

    block_args = dict(
        block_type=block_type,
        out_chs=int(options["c"]),
        stride=int(options["s"]),
        act_layer=act_layer,
    )

    if block_type == "ir":
        block_args.update(
            dict(
                dw_kernel_size=parse_ksize(options["k"]),
                exp_ratio=float(options["e"]),
                se_ratio=float(options.get("se", 0.0)),
                noskip=skip is False,
            )
        )
    elif block_type == "uir":
        start_kernel_size = parse_ksize(options.get("a", "0"))
        end_kernel_size = parse_ksize(options.get("p", "0"))
        block_args.update(
            dict(
                dw_kernel_size_start=start_kernel_size,
                dw_kernel_size_mid=parse_ksize(options["k"]),
                dw_kernel_size_end=end_kernel_size,
                exp_ratio=float(options["e"]),
                se_ratio=float(options.get("se", 0.0)),
                noskip=skip is False,
            )
        )
    elif block_type == "er":
        block_args.update(
            dict(
                exp_kernel_size=parse_ksize(options["k"]),
                pw_kernel_size=1,
                exp_ratio=float(options["e"]),
                se_ratio=float(options.get("se", 0.0)),
                noskip=skip is False,
            )
        )
    elif block_type in ("mqa", "mha"):
        key_dim_val = int(options.get("d", "64"))
        block_args.update(
            dict(
                num_heads=int(options.get("h", "12")),
                key_dim=key_dim_val,
                value_dim=key_dim_val,
                kv_stride=int(options.get("v", "1")),
                use_cpe=bool(int(options.get("cpe", "0"))),
            )
        )
    return block_args, num_repeat


def decode_arch_def(arch_def):
    arch_args = []
    for _, block_strings in enumerate(arch_def):
        stack_args = []
        for block_str in block_strings:
            ba, rep = decode_block_str(block_str)
            stack_args.extend([deepcopy(ba) for _ in range(rep)])
        arch_args.append(stack_args)
    return arch_args


def convert_arch_def_to_stackwise(arch_def):
    decoded_args = decode_arch_def(arch_def)
    stackwise_params = {
        k: []
        for k in [
            "stackwise_block_types",
            "stackwise_num_blocks",
            "stackwise_num_filters",
            "stackwise_strides",
            "stackwise_act_layers",
            "stackwise_exp_ratios",
            "stackwise_se_ratios",
            "stackwise_dw_kernel_sizes",
            "stackwise_dw_start_kernel_sizes",
            "stackwise_dw_end_kernel_sizes",
            "stackwise_exp_kernel_sizes",
            "stackwise_pw_kernel_sizes",
            "stackwise_num_heads",
            "stackwise_key_dims",
            "stackwise_value_dims",
            "stackwise_kv_strides",
            "stackwise_use_cpe",
        ]
    }
    for stack in decoded_args:
        stackwise_params["stackwise_num_blocks"].append(len(stack))
        current_stack_params = {
            k: [] for k in stackwise_params if k != "stackwise_num_blocks"
        }
        for block in stack:
            current_stack_params["stackwise_block_types"].append(
                block.get("block_type")
            )
            current_stack_params["stackwise_num_filters"].append(
                block.get("out_chs")
            )
            current_stack_params["stackwise_strides"].append(
                block.get("stride")
            )
            current_stack_params["stackwise_act_layers"].append(
                block.get("act_layer")
            )
            current_stack_params["stackwise_exp_ratios"].append(
                block.get("exp_ratio", 0.0)
            )
            current_stack_params["stackwise_se_ratios"].append(
                block.get("se_ratio", 0.0)
            )
            current_stack_params["stackwise_dw_kernel_sizes"].append(
                block.get("dw_kernel_size", block.get("dw_kernel_size_mid", 0))
            )
            current_stack_params["stackwise_dw_start_kernel_sizes"].append(
                block.get("dw_kernel_size_start", 0)
            )
            current_stack_params["stackwise_dw_end_kernel_sizes"].append(
                block.get("dw_kernel_size_end", 0)
            )
            current_stack_params["stackwise_exp_kernel_sizes"].append(
                block.get("exp_kernel_size", 0)
            )
            current_stack_params["stackwise_pw_kernel_sizes"].append(
                block.get("pw_kernel_size", 0)
            )
            current_stack_params["stackwise_num_heads"].append(
                block.get("num_heads", 0)
            )
            current_stack_params["stackwise_key_dims"].append(
                block.get("key_dim", 0)
            )
            current_stack_params["stackwise_value_dims"].append(
                block.get("value_dim", 0)
            )
            current_stack_params["stackwise_kv_strides"].append(
                block.get("kv_stride", 0)
            )
            current_stack_params["stackwise_use_cpe"].append(
                block.get("use_cpe", False)
            )
        for key, value in current_stack_params.items():
            stackwise_params[key].append(value)
    return stackwise_params


class MobileNetV5Builder:
    """Builds a MobileNetV5 model from a decoded architecture definition.

    This class takes a decoded architecture definition and constructs a list of
    network stages, where each stage is a list of blocks. It handles channel
    rounding, stride management, and feature extraction points.

    Args:
        output_stride: int. The desired output stride of the network.
        pad_type: str. The padding type for convolutions.
        round_chs_fn: callable. A function to round the number of channels.
        se_from_exp: bool. If `True`, SE channel reduction is based on the
            expanded channels.
        act_layer: str. The default activation function for blocks.
        norm_layer: str. The default normalization layer for blocks.
        aa_layer: keras.layers.Layer. An optional anti-aliasing layer.
        se_layer: keras.layers.Layer. The Squeeze-and-Excitation layer to use.
        drop_path_rate: float. The stochastic depth rate for the network.
        layer_scale_init_value: float. The initial value for layer scale.
        feature_location: str. Where to extract features from, either
            `"bottleneck"`, `"expansion"`, or `""`.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        output_stride=32,
        pad_type="same",
        round_chs_fn=round_channels,
        se_from_exp=False,
        act_layer="relu",
        norm_layer="batch_norm",
        aa_layer=None,
        se_layer=None,
        drop_path_rate=0.0,
        layer_scale_init_value=None,
        feature_location="",
        data_format=None,
        channel_axis=None,
        dtype=None,
    ):
        self.output_stride = output_stride
        self.pad_type = pad_type
        self.data_format = data_format
        self.channel_axis = channel_axis
        self.round_chs_fn = round_chs_fn
        self.se_from_exp = se_from_exp
        self.act_layer = act_layer
        self.norm_layer = norm_layer
        self.aa_layer = aa_layer
        self.se_layer = se_layer
        self.drop_path_rate = drop_path_rate
        self.layer_scale_init_value = layer_scale_init_value
        self.dtype = dtype
        if feature_location == "depthwise":
            feature_location = "expansion"
        self.feature_location = feature_location
        assert feature_location in ("bottleneck", "expansion", "")
        self.in_chs = None
        self.features = []

    def _make_block(self, ba, block_idx, block_count):
        drop_path_rate = self.drop_path_rate * block_idx / block_count
        bt = ba.pop("block_type")
        ba["filters"] = self.round_chs_fn(ba.pop("out_chs"))
        s2d = ba.get("s2d", 0)
        if s2d > 0:
            ba["filters"] *= 4
        if "expansion_in_chs" in ba and ba["expansion_in_chs"]:
            ba["expansion_in_chs"] = self.round_chs_fn(ba["expansion_in_chs"])
        ba["pad_type"] = self.pad_type
        ba["act_layer"] = (
            ba.get("act_layer")
            if ba.get("act_layer") is not None
            else self.act_layer
        )
        assert ba["act_layer"] is not None
        ba["norm_layer"] = self.norm_layer
        ba["drop_path_rate"] = drop_path_rate
        ba["data_format"] = self.data_format
        ba["channel_axis"] = self.channel_axis
        ba["dtype"] = self.dtype
        if bt in ("ir", "er", "uir", "ds", "dsa"):
            se_ratio = ba.pop("se_ratio", None)
            if se_ratio and self.se_layer is not None:
                if not self.se_from_exp:
                    se_ratio /= ba.get("exp_ratio", 1.0)
                if s2d == 1:
                    se_ratio /= 4
                ba["se_layer"] = lambda channels: self.se_layer(
                    filters=channels,
                    bottleneck_filters=adjust_channels(channels * se_ratio),
                    squeeze_activation=ba["act_layer"],
                    excite_activation="sigmoid",
                    data_format=self.data_format,
                    dtype=self.dtype,
                )
            else:
                ba["se_layer"] = None
        ba.pop("aa_layer", None)
        if bt == "ir":
            padding = 0
            if ba["pad_type"].lower() in ("", "same"):
                kernel_size = ba["dw_kernel_size"]
                if isinstance(kernel_size, (list, tuple)):
                    kernel_size = kernel_size[0]
                padding = (kernel_size - 1) // 2
            block = (
                CondConvResidual(**ba)
                if ba.get("num_experts", 0) > 0
                else InvertedResidualBlock(
                    expansion=ba["exp_ratio"],
                    infilters=self.in_chs,
                    filters=ba["filters"],
                    kernel_size=ba["dw_kernel_size"],
                    stride=ba["stride"],
                    padding=padding,
                    squeeze_excite_ratio=ba.pop("se_ratio", None),
                    activation=ba["act_layer"],
                )
            )
        elif bt == "ds" or bt == "dsa":
            block = DepthwiseConvBlock(
                infilters=self.in_chs,
                filters=ba["filters"],
                kernel_size=ba["dw_kernel_size"],
                stride=ba["stride"],
                squeeze_excite_ratio=ba.pop("se_ratio", None),
                residual=not ba["noskip"],
                dtype=self.dtype,
            )
        elif bt == "er":
            block = EdgeResidual(**ba)
        elif bt == "cn":
            block = ConvBnActBlock(out_chs=ba.pop("filters"), **ba)
        elif bt == "uir":
            block = UniversalInvertedResidual(
                **ba, layer_scale_init_value=self.layer_scale_init_value
            )
        elif bt == "mqa":
            ba.pop("act_layer", None)
            block = MobileAttention(
                **ba,
                use_multi_query=True,
                layer_scale_init_value=self.layer_scale_init_value,
            )
        elif bt == "mha":
            ba.pop("act_layer", None)
            block = MobileAttention(
                **ba, layer_scale_init_value=self.layer_scale_init_value
            )
        else:
            raise ValueError(f"Unknown block type ({bt}) while building model.")
        self.in_chs = ba["filters"]
        return block

    def __call__(self, in_chs, model_block_args):
        self.in_chs = in_chs
        total_block_count = sum([len(x) for x in model_block_args])
        total_block_idx = 0
        current_stride = 2
        current_dilation = 1
        stages = []
        if model_block_args[0][0]["stride"] > 1:
            feature_info = dict(
                module="conv_stem",
                num_chs=in_chs,
                stage=0,
                reduction=current_stride,
            )
            self.features.append(feature_info)
        space2depth = 0
        for stack_idx, stack_args in enumerate(model_block_args):
            blocks = []
            for block_idx, block_args in enumerate(stack_args):
                last_block = block_idx + 1 == len(stack_args)
                in_chs_for_current_block = self.in_chs
                assert block_args["stride"] in (1, 2)
                if block_idx >= 1:
                    block_args["stride"] = 1
                if not space2depth and block_args.pop("s2d", False):
                    assert block_args["stride"] == 1
                    space2depth = 1
                if space2depth > 0:
                    if space2depth == 2 and block_args["stride"] == 2:
                        block_args["stride"] = 1
                        block_args["exp_ratio"] /= 4
                        space2depth = 0
                    else:
                        block_args["s2d"] = space2depth
                next_dilation = current_dilation
                if block_args["stride"] > 1:
                    next_output_stride = current_stride * block_args["stride"]
                    if next_output_stride > self.output_stride:
                        next_dilation = current_dilation * block_args["stride"]
                        block_args["stride"] = 1
                    else:
                        current_stride = next_output_stride
                block_args["dilation"] = current_dilation
                if next_dilation != current_dilation:
                    current_dilation = next_dilation
                block = self._make_block(
                    block_args.copy(), total_block_idx, total_block_count
                )
                blocks.append(block)
                if space2depth == 1:
                    space2depth = 2
                extract_features = False
                if last_block:
                    next_stack_idx = stack_idx + 1
                    extract_features = (
                        next_stack_idx >= len(model_block_args)
                        or model_block_args[next_stack_idx][0]["stride"] > 1
                    )
                if extract_features:
                    num_chs = 0
                    module_name = f"blocks.{stack_idx}.{block_idx}"
                    if self.feature_location == "expansion":
                        bt = block_args.get("block_type")
                        if bt in ["ir", "er", "uir"]:
                            exp_ratio = block_args.get("exp_ratio", 1.0)
                            num_chs = self.round_chs_fn(
                                in_chs_for_current_block * exp_ratio
                            )
                        else:
                            num_chs = in_chs_for_current_block
                    else:
                        num_chs = self.in_chs
                        module_name = f"blocks.{stack_idx}"

                    feature_info = dict(
                        stage=stack_idx + 1,
                        reduction=current_stride,
                        num_chs=num_chs,
                        module=module_name,
                    )
                    self.features.append(feature_info)
                total_block_idx += 1
            stages.append(blocks)
        return stages



import json
import os
import shutil
import warnings

import keras

from keras_hub.src.utils.transformers.export.gemma import get_gemma_config
from keras_hub.src.utils.transformers.export.gemma import (
    get_gemma_tokenizer_config,
)
from keras_hub.src.utils.transformers.export.gemma import get_gemma_weights_map

MODEL_CONFIGS = {
    "GemmaBackbone": get_gemma_config,
    # Add for future models, e.g., "MistralBackbone": get_mistral_config
}

MODEL_EXPORTERS = {
    "GemmaBackbone": get_gemma_weights_map,
    # Add for future models, e.g., "MistralBackbone": get_mistral_weights_map
}

MODEL_TOKENIZER_CONFIGS = {
    "GemmaTokenizer": get_gemma_tokenizer_config,
    # Add for future models, e.g., "MistralTokenizer":
    # get_mistral_tokenizer_config
}


def export_backbone(backbone, path, include_lm_head=False):
    """Export the backbone model to HuggingFace format.

    Args:
        backbone: The Keras backbone model to convert.
        path: str. Path to save the exported model.
        include_lm_head: bool. If True, include lm_head weights if applicable.
    """
    backend = keras.config.backend()
    model_type = backbone.__class__.__name__
    if model_type not in MODEL_CONFIGS:
        raise ValueError(
            f"Export to Transformers format not implemented for {model_type}"
        )
    if model_type not in MODEL_EXPORTERS:
        raise ValueError(
            f"Export to Transformers format not implemented for {model_type}"
        )
    # Get config
    get_config_fn = MODEL_CONFIGS[model_type]
    hf_config = get_config_fn(backbone)
    # Get weights
    get_weights_fn = MODEL_EXPORTERS[model_type]
    weights_dict = get_weights_fn(backbone, include_lm_head=include_lm_head)
    if not weights_dict:
        raise ValueError("No weights to save.")
    # Save config
    os.makedirs(path, exist_ok=True)
    config_path = os.path.join(path, "config.json")
    with open(config_path, "w") as f:
        json.dump(hf_config, f)
    # Save weights based on backend
    weights_path = os.path.join(path, "model.safetensors")
    if backend == "torch":
        from safetensors.torch import save_file

        weights_dict_contiguous = {
            k: v.value.contiguous() if hasattr(v, "value") else v.contiguous()
            for k, v in weights_dict.items()
        }
        save_file(
            weights_dict_contiguous, weights_path, metadata={"format": "pt"}
        )
    elif backend == "tensorflow":
        from safetensors.tensorflow import save_file

        save_file(weights_dict, weights_path, metadata={"format": "pt"})
    elif backend == "jax":
        from safetensors.flax import save_file

        save_file(weights_dict, weights_path, metadata={"format": "pt"})
    else:
        raise ValueError(f"Unsupported backend: {backend}")


def export_tokenizer(tokenizer, path):
    """Export only the tokenizer to HuggingFace Transformers format.

    Args:
        tokenizer: The Keras tokenizer to convert.
        path: str. Path to save the exported tokenizer.
    """
    os.makedirs(path, exist_ok=True)
    # Save tokenizer assets
    tokenizer.save_assets(path)
    # Export tokenizer config
    tokenizer_type = tokenizer.__class__.__name__
    if tokenizer_type not in MODEL_TOKENIZER_CONFIGS:
        raise ValueError(
            "Export to Transformers format not implemented for {tokenizer_type}"
        )
    get_tokenizer_config_fn = MODEL_TOKENIZER_CONFIGS[tokenizer_type]
    tokenizer_config = get_tokenizer_config_fn(tokenizer)
    tokenizer_config_path = os.path.join(path, "tokenizer_config.json")
    with open(tokenizer_config_path, "w") as f:
        json.dump(tokenizer_config, f, indent=4)
    # Rename vocabulary file
    vocab_spm_path = os.path.join(path, "vocabulary.spm")
    tokenizer_model_path = os.path.join(path, "tokenizer.model")
    if os.path.exists(vocab_spm_path):
        shutil.move(vocab_spm_path, tokenizer_model_path)
    else:
        warnings.warn(
            f"{vocab_spm_path} not found. Tokenizer may not load "
            "correctly. Ensure that the tokenizer configuration "
            "is correct and that the vocabulary file is present "
            "in the original model."
        )


def export_to_safetensors(keras_model, path):
    """Converts a Keras model to Hugging Face Transformers format.

    It does the following:
    - Exports the backbone (config and weights).
    - Exports the tokenizer assets.

    Args:
        keras_model: The Keras model to convert.
        path: str. Path of the directory to which the safetensors file,
          config and tokenizer will be saved.
    """
    backbone = keras_model.backbone
    export_backbone(backbone, path, include_lm_head=True)
    if (
        keras_model.preprocessor is not None
        and keras_model.preprocessor.tokenizer is None
    ):
        raise ValueError(
            "CausalLM preprocessor must have a tokenizer for export "
            "if attached."
        )
    if keras_model.preprocessor is not None:
        export_tokenizer(keras_model.preprocessor.tokenizer, path)



from keras_hub.src.models.distil_bert.distil_bert_backbone import (
    DistilBertBackbone,
)
from keras_hub.src.models.distil_bert.distil_bert_presets import (
    backbone_presets,
)
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, DistilBertBackbone)



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.start_end_packer import StartEndPacker
from keras_hub.src.models.causal_lm_preprocessor import CausalLMPreprocessor
from keras_hub.src.models.clip.clip_backbone import CLIPBackbone
from keras_hub.src.models.clip.clip_image_converter import CLIPImageConverter
from keras_hub.src.models.clip.clip_tokenizer import CLIPTokenizer
from keras_hub.src.utils.tensor_utils import preprocessing_function

try:
    import tensorflow as tf
except ImportError:
    tf = None


@keras_hub_export("keras_hub.models.CLIPPreprocessor")
class CLIPPreprocessor(CausalLMPreprocessor):
    """CLIP preprocessor.

    This preprocessing layer will do 2 things:

    This preprocessing layer is meant for use with
    `keras_hub.models.CLIPBackbone`. By default, it will take in batches of
    strings and images, and return token ids and resized images.

    Args:
        tokenizer: A `keras_hub.models.CLIPTokenizer` instance.
        image_converter: A `keras_hub.models.CLIPImageConverter` instance.
        sequence_length: The length of the packed inputs.
        add_start_token: If `True`, the preprocessor will prepend the tokenizer
            start token to each input sequence.
        add_end_token: If `True`, the preprocessor will append the tokenizer
            end token to each input sequence.
        to_lower: bool. Whether to lower the inputs.

    Call arguments:
        x: A dict with `"prompts"` and `"images"` keys, where `"prompts"` is
            `tf.Tensor` or list of python strings and `"images"` are the image
            tensors.
        y: Label data. Should always be `None` since SigLIP doesn't need the
            label to calculate the loss.
        sample_weight: Label weights.
        sequence_length: Pass to override the configured `sequence_length` of
            the layer.

    Examples:
    ```python
    # Load the preprocessor from a preset.
    preprocessor = keras_hub.models.CLIPPreprocessor.from_preset(
        "clip_vit_base_patch16"
    )

    # Tokenize the sentence and preprocess the image.
    preprocessor(
        {
            "prompts": "The quick brown fox jumped.",
            "images": np.ones(shape=(123, 123, 3)),
        }
    )

    # Tokenize a batch of sentences and preprocess a batch of images.
    preprocessor(
        {
            "prompts": ["The quick brown fox jumped.", "The fox slept."],
            "images": np.ones(shape=(2, 123, 123, 3)),
        }
    )
    ```
    """

    backbone_cls = CLIPBackbone
    tokenizer_cls = CLIPTokenizer
    image_converter_cls = CLIPImageConverter

    def __init__(
        self,
        tokenizer,
        image_converter=None,
        sequence_length=77,
        add_start_token=True,
        add_end_token=True,
        to_lower=True,
        **kwargs,
    ):
        super().__init__(
            tokenizer=tokenizer,
            sequence_length=sequence_length,
            add_start_token=add_start_token,
            add_end_token=add_end_token,
            **kwargs,
        )
        self.image_converter = image_converter
        self.to_lower = to_lower

    def build(self, input_shape):
        # Defer packer creation to `build()` so that we can be sure tokenizer
        # assets have loaded when restoring a saved model.
        self.packer = StartEndPacker(
            start_value=self.tokenizer.start_token_id,
            end_value=self.tokenizer.end_token_id,
            pad_value=self.tokenizer.pad_token_id,
            sequence_length=self.sequence_length,
            return_padding_mask=True,
        )
        self.built = True

    @preprocessing_function
    def call(
        self,
        x,
        y=None,
        sample_weight=None,
        sequence_length=None,
    ):
        sequence_length = sequence_length or self.sequence_length
        images, prompts = x["images"], x["prompts"]
        if self.to_lower:
            prompts = tf.strings.lower(prompts)
        prompts = self.tokenizer(prompts)
        if images is not None and self.image_converter:
            images = self.image_converter(images)
        token_ids, padding_mask = self.packer(
            prompts,
            sequence_length=sequence_length,
            add_start_value=self.add_start_token,
            add_end_value=self.add_end_token,
        )
        x = {
            "token_ids": token_ids,
            "padding_mask": padding_mask,
            "images": images,
        }
        return keras.utils.pack_x_y_sample_weight(x, y, sample_weight)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "to_lower": self.to_lower,
            }
        )
        return config



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.utils.keras_utils import clone_initializer

from keras_hub.src.layers.modeling.transformer_layer_utils import (  # isort:skip
    merge_padding_and_attention_mask,
)


@keras_hub_export("keras_hub.layers.TransformerEncoder")
class TransformerEncoder(keras.layers.Layer):
    """Transformer encoder.

    This class follows the architecture of the transformer encoder layer in the
    paper [Attention is All You Need](https://arxiv.org/abs/1706.03762). Users
    can instantiate multiple instances of this class to stack up an encoder.

    This layer will compute an attention mask, prioritizing explicitly provided
    masks (a `padding_mask` or a custom `attention_mask`) over an implicit Keras
    padding mask (for example, by passing `mask_zero=True` to a
    `keras.layers.Embedding` layer). If both a `padding_mask` and a
    `attention_mask` are provided, they will be combined to determine the final
    mask. See the Masking and Padding
    [guide](https://keras.io/guides/understanding_masking_and_padding/)
    for more details.

    Args:
        intermediate_dim: int, the hidden size of feedforward network.
        num_heads: int, the number of heads in the
            `keras.layers.MultiHeadAttention` layer.
        dropout: float. the dropout value, shared by
            `keras.layers.MultiHeadAttention` and feedforward network.
            Defaults to `0.`.
        activation: string or `keras.activations`. the
            activation function of feedforward network.
            Defaults to `"relu"`.
        layer_norm_epsilon: float. The epsilon value in layer
            normalization components. Defaults to `1e-5`.
        kernel_initializer: string or `keras.initializers` initializer.
            The kernel initializer for the dense and multiheaded
            attention layers. Defaults to `"glorot_uniform"`.
        bias_initializer: string or `keras.initializers` initializer.
            The bias initializer for the dense and multiheaded
            attention layers. Defaults to `"zeros"`.
        normalize_first: bool. If True, the inputs to the
            attention layer and the intermediate dense layer  are normalized
            (similar to GPT-2). If set to False, outputs of attention layer and
            intermediate dense layer are normalized (similar to BERT).
            Defaults to `False`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `trainable`, `dtype` etc.

    Example:

    ```python
    # Create a single transformer encoder layer.
    encoder = keras_hub.layers.TransformerEncoder(
        intermediate_dim=64, num_heads=8)

    # Create a simple model containing the encoder.
    input = keras.Input(shape=(10, 64))
    output = encoder(input)
    model = keras.Model(inputs=input, outputs=output)

    # Call encoder on the inputs.
    input_data = np.random.uniform(size=(2, 10, 64))
    output = model(input_data)
    ```

    References:
     - [Vaswani et al., 2017](https://arxiv.org/abs/1706.03762)
    """

    def __init__(
        self,
        intermediate_dim,
        num_heads,
        dropout=0,
        activation="relu",
        layer_norm_epsilon=1e-05,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        normalize_first=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.intermediate_dim = intermediate_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.activation = keras.activations.get(activation)
        self.layer_norm_epsilon = layer_norm_epsilon
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.normalize_first = normalize_first
        self.supports_masking = True

    def build(self, inputs_shape):
        # Infer the dimension of our hidden feature size from the build shape.
        hidden_dim = inputs_shape[-1]
        # Attention head size is `hidden_dim` over the number of heads.
        key_dim = int(hidden_dim // self.num_heads)
        if key_dim == 0:
            raise ValueError(
                "Attention `key_dim` computed cannot be zero. "
                f"The `hidden_dim` value of {hidden_dim} has to be equal to "
                f"or greater than `num_heads` value of {self.num_heads}."
            )

        # Self attention layers.
        self._self_attention_layer = keras.layers.MultiHeadAttention(
            num_heads=self.num_heads,
            key_dim=key_dim,
            dropout=self.dropout,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
            dtype=self.dtype_policy,
            name="self_attention_layer",
        )
        if hasattr(self._self_attention_layer, "_build_from_signature"):
            self._self_attention_layer._build_from_signature(
                query=inputs_shape,
                value=inputs_shape,
            )
        else:
            self._self_attention_layer.build(
                query_shape=inputs_shape,
                value_shape=inputs_shape,
            )
        self._self_attention_layer_norm = keras.layers.LayerNormalization(
            epsilon=self.layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="self_attention_layer_norm",
        )
        self._self_attention_layer_norm.build(inputs_shape)
        self._self_attention_dropout = keras.layers.Dropout(
            rate=self.dropout,
            dtype=self.dtype_policy,
            name="self_attention_dropout",
        )

        # Feedforward layers.
        self._feedforward_layer_norm = keras.layers.LayerNormalization(
            epsilon=self.layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="feedforward_layer_norm",
        )
        self._feedforward_layer_norm.build(inputs_shape)
        self._feedforward_intermediate_dense = keras.layers.Dense(
            self.intermediate_dim,
            activation=self.activation,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
            dtype=self.dtype_policy,
            name="feedforward_intermediate_dense",
        )
        self._feedforward_intermediate_dense.build(inputs_shape)
        self._feedforward_output_dense = keras.layers.Dense(
            hidden_dim,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
            dtype=self.dtype_policy,
            name="feedforward_output_dense",
        )
        intermediate_shape = list(inputs_shape)
        intermediate_shape[-1] = self.intermediate_dim
        self._feedforward_output_dense.build(tuple(intermediate_shape))
        self._feedforward_dropout = keras.layers.Dropout(
            rate=self.dropout,
            dtype=self.dtype_policy,
            name="feedforward_dropout",
        )
        self.built = True

    def call(
        self,
        inputs,
        padding_mask=None,
        attention_mask=None,
        training=None,
        return_attention_scores=False,
    ):
        """Forward pass of the TransformerEncoder.

        Args:
            inputs: a Tensor. The input data to TransformerEncoder, should be
                of shape [batch_size, sequence_length, hidden_dim].
            padding_mask: a boolean Tensor. It indicates if the token should be
                masked because the token is introduced due to padding.
                `padding_mask` should have shape [batch_size, sequence_length].
            attention_mask: a boolean Tensor. Customized mask used to mask out
                certain tokens. `attention_mask` should have shape
                [batch_size, sequence_length, sequence_length].
            training: a boolean indicating whether the layer should behave in
                training mode or in inference mode.
            return_attention_scores: a boolean indicating whether the output
                should be `(attention_output, attention_scores)` if `True` or
                `attention_output` if `False`. Defaults to `False`.

        Returns:
            A Tensor of the same shape as the `inputs`.
        """
        x = inputs  # Intermediate result.

        # Compute self attention mask.
        self_attention_mask = merge_padding_and_attention_mask(
            inputs, padding_mask, attention_mask
        )

        # Self attention block.
        residual = x
        if self.normalize_first:
            x = self._self_attention_layer_norm(x)

        if return_attention_scores:
            x, attention_scores = self._self_attention_layer(
                query=x,
                value=x,
                attention_mask=self_attention_mask,
                return_attention_scores=return_attention_scores,
                training=training,
            )
        else:
            x = self._self_attention_layer(
                query=x,
                value=x,
                attention_mask=self_attention_mask,
                training=training,
            )

        x = self._self_attention_dropout(x, training=training)
        x = x + residual
        if not self.normalize_first:
            x = self._self_attention_layer_norm(x)

        # Feedforward block.
        residual = x
        if self.normalize_first:
            x = self._feedforward_layer_norm(x)
        x = self._feedforward_intermediate_dense(x)
        x = self._feedforward_output_dense(x)
        x = self._feedforward_dropout(x, training=training)
        x = x + residual
        if not self.normalize_first:
            x = self._feedforward_layer_norm(x)

        if return_attention_scores:
            return x, attention_scores

        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "intermediate_dim": self.intermediate_dim,
                "num_heads": self.num_heads,
                "dropout": self.dropout,
                "activation": keras.activations.serialize(self.activation),
                "layer_norm_epsilon": self.layer_norm_epsilon,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "normalize_first": self.normalize_first,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        return inputs_shape



import keras

from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import DropPath
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import LayerScale2d
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import RmsNorm2d


class MultiQueryAttention2d(keras.layers.Layer):
    """Implements 2D Multi-Query Attention.

    This layer performs attention on 2D spatial inputs. It uses a multi-query
    attention mechanism where multiple query heads attend to a single key and
    value.

    Args:
        filters: int. The output channel dimension.
        num_heads: int. The number of attention heads.
        key_dim: int. The dimension of the key. If `None`, it is calculated as
            `dim // num_heads`.
        value_dim: int. The dimension of the value. If `None`, it is calculated
            as `dim // num_heads`.
        query_strides: int or tuple. The stride for downsampling the query.
        kv_stride: int. The stride for downsampling the key and value.
        dw_kernel_size: int. The kernel size for the depthwise convolution used
            for downsampling.
        dilation: int. The dilation rate for the depthwise convolution.
        padding: str. The padding type for convolutions.
        attn_drop: float. The dropout rate for the attention weights.
        proj_drop: float. The dropout rate for the output projection.
        norm_layer: keras.layers.Layer. The normalization layer to use.
        use_bias: bool. If `True`, bias terms are used in convolutions.
        channel_axis: int. The axis representing the channels in the input
            tensor.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
    """

    def __init__(
        self,
        filters,
        num_heads=8,
        key_dim=None,
        value_dim=None,
        query_strides=1,
        kv_stride=1,
        dw_kernel_size=3,
        dilation=1,
        padding="same",
        attn_drop=0.0,
        proj_drop=0.0,
        norm_layer=keras.layers.BatchNormalization,
        use_bias=False,
        channel_axis=None,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.num_heads = num_heads
        self.key_dim_arg = key_dim
        self.value_dim_arg = value_dim
        self.query_strides_arg = query_strides
        self.kv_stride = kv_stride
        self.dw_kernel_size = dw_kernel_size
        self.dilation = dilation
        self.padding_arg = padding
        self.attn_drop_rate = attn_drop
        self.proj_drop_rate = proj_drop
        self.norm_layer = norm_layer
        self.use_bias = use_bias
        self.channel_axis = channel_axis
        self.data_format = data_format
        self.query_strides = (
            query_strides
            if isinstance(query_strides, (list, tuple))
            else (query_strides, query_strides)
        )
        self.has_query_strides = any([s > 1 for s in self.query_strides])
        self.padding = padding
        self.conv_kernel_initializer = keras.initializers.VarianceScaling(
            scale=2.0, mode="fan_out", distribution="untruncated_normal"
        )
        self.bias_initializer = "zeros"
        self.attn_drop_layer = keras.layers.Dropout(
            attn_drop, dtype=self.dtype_policy
        )

    def build(self, input_shape):
        super().build(input_shape)
        dim = input_shape[self.channel_axis]
        self.key_dim = self.key_dim_arg or dim // self.num_heads
        self.value_dim = self.value_dim_arg or dim // self.num_heads
        self.scale = self.key_dim**-0.5
        query_layers = []
        if self.has_query_strides:
            pool_padding = "valid" if self.padding == "valid" else "same"
            query_layers.append(
                keras.layers.AveragePooling2D(
                    pool_size=self.query_strides,
                    strides=self.query_strides,
                    padding=pool_padding,
                    data_format=self.data_format,
                    name="query_down_pool",
                    dtype=self.dtype_policy,
                )
            )
            if self.norm_layer is RmsNorm2d:
                norm = self.norm_layer(
                    dim=dim,
                    channel_axis=self.channel_axis,
                    data_format=self.data_format,
                    name="query_norm",
                    dtype=self.dtype_policy,
                )
            else:
                norm = self.norm_layer(
                    axis=self.channel_axis,
                    name="query_norm",
                    gamma_initializer="ones",
                    beta_initializer="zeros",
                    dtype=self.dtype_policy,
                )
            query_layers.append(norm)
        query_layers.append(
            keras.layers.Conv2D(
                filters=self.num_heads * self.key_dim,
                kernel_size=1,
                use_bias=self.use_bias,
                data_format=self.data_format,
                name="query_proj",
                kernel_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
        )
        self.query_layers = query_layers
        key_layers = []
        if self.kv_stride > 1:
            key_layers.append(
                keras.layers.DepthwiseConv2D(
                    kernel_size=self.dw_kernel_size,
                    strides=self.kv_stride,
                    dilation_rate=self.dilation,
                    padding=self.padding,
                    data_format=self.data_format,
                    name="key_down_conv",
                    depthwise_initializer=self.conv_kernel_initializer,
                    bias_initializer=self.bias_initializer,
                    use_bias=False,
                    dtype=self.dtype_policy,
                )
            )
            if self.norm_layer is RmsNorm2d:
                norm = self.norm_layer(
                    dim=dim,
                    channel_axis=self.channel_axis,
                    data_format=self.data_format,
                    name="key_norm",
                    dtype=self.dtype_policy,
                )
            else:
                norm = self.norm_layer(
                    axis=self.channel_axis,
                    gamma_initializer="ones",
                    beta_initializer="zeros",
                    name="key_norm",
                    dtype=self.dtype_policy,
                )
            key_layers.append(norm)
        key_layers.append(
            keras.layers.Conv2D(
                filters=self.key_dim,
                kernel_size=1,
                padding="valid",
                use_bias=self.use_bias,
                data_format=self.data_format,
                name="key_proj",
                kernel_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
        )
        self.key_layers = key_layers
        value_layers = []
        if self.kv_stride > 1:
            value_layers.append(
                keras.layers.DepthwiseConv2D(
                    kernel_size=self.dw_kernel_size,
                    strides=self.kv_stride,
                    dilation_rate=self.dilation,
                    padding=self.padding,
                    data_format=self.data_format,
                    name="value_down_conv",
                    depthwise_initializer=self.conv_kernel_initializer,
                    bias_initializer=self.bias_initializer,
                    use_bias=False,
                    dtype=self.dtype_policy,
                )
            )
            if self.norm_layer is RmsNorm2d:
                norm = self.norm_layer(
                    dim=dim,
                    channel_axis=self.channel_axis,
                    data_format=self.data_format,
                    name="value_norm",
                    dtype=self.dtype_policy,
                )
            else:
                norm = self.norm_layer(
                    axis=self.channel_axis,
                    gamma_initializer="ones",
                    beta_initializer="zeros",
                    name="value_norm",
                    dtype=self.dtype_policy,
                )
            value_layers.append(norm)
        value_layers.append(
            keras.layers.Conv2D(
                filters=self.value_dim,
                kernel_size=1,
                padding="valid",
                use_bias=self.use_bias,
                data_format=self.data_format,
                name="value_proj",
                kernel_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
        )
        self.value_layers = value_layers
        output_layers = []
        if self.has_query_strides:
            output_layers.append(
                keras.layers.UpSampling2D(
                    size=self.query_strides,
                    interpolation="bilinear",
                    data_format=self.data_format,
                    name="output_upsample",
                    dtype=self.dtype_policy,
                )
            )
        output_layers.append(
            keras.layers.Conv2D(
                filters=self.filters,
                kernel_size=1,
                use_bias=self.use_bias,
                data_format=self.data_format,
                name="output_proj",
                kernel_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
        )
        output_layers.append(
            keras.layers.Dropout(self.proj_drop_rate, dtype=self.dtype_policy)
        )
        self.output_proj_layers = output_layers

    def call(self, x, training=False):
        B = keras.ops.shape(x)[0]
        q = x
        for layer in self.query_layers:
            try:
                q = layer(q, training=training)
            except TypeError:
                q = layer(q)
        k = x
        for layer in self.key_layers:
            try:
                k = layer(k, training=training)
            except TypeError:
                k = layer(k)
        v = x
        for layer in self.value_layers:
            try:
                v = layer(v, training=training)
            except TypeError:
                v = layer(v)
        if self.data_format == "channels_last":
            q = keras.ops.transpose(q, (0, 3, 1, 2))
            k = keras.ops.transpose(k, (0, 3, 1, 2))
            v = keras.ops.transpose(v, (0, 3, 1, 2))
        s_q = keras.ops.shape(q)
        h_q, w_q = s_q[2], s_q[3]
        q = keras.ops.reshape(q, (B, self.num_heads, self.key_dim, -1))
        q = keras.ops.transpose(q, (0, 1, 3, 2))
        k = keras.ops.reshape(k, (B, self.key_dim, -1))
        k = keras.ops.transpose(k, (0, 2, 1))
        k = keras.ops.expand_dims(k, axis=1)
        v = keras.ops.reshape(v, (B, self.value_dim, -1))
        v = keras.ops.transpose(v, (0, 2, 1))
        v = keras.ops.expand_dims(v, axis=1)
        q = q * self.scale
        attn = keras.ops.matmul(q, keras.ops.transpose(k, (0, 1, 3, 2)))
        attn = keras.ops.softmax(attn, axis=-1)
        attn = self.attn_drop_layer(attn, training=training)
        o = keras.ops.matmul(attn, v)
        o = keras.ops.transpose(o, (0, 2, 1, 3))
        feat_dim = self.num_heads * self.value_dim
        o = keras.ops.reshape(o, (B, h_q, w_q, feat_dim))
        if self.data_format == "channels_first":
            o = keras.ops.transpose(o, (0, 3, 1, 2))
        x_out = o
        for layer in self.output_proj_layers:
            try:
                x_out = layer(x_out, training=training)
            except TypeError:
                x_out = layer(x_out)
        return x_out

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "num_heads": self.num_heads,
                "key_dim": self.key_dim_arg,
                "value_dim": self.value_dim_arg,
                "query_strides": self.query_strides_arg,
                "kv_stride": self.kv_stride,
                "dw_kernel_size": self.dw_kernel_size,
                "dilation": self.dilation,
                "padding": self.padding_arg,
                "attn_drop": self.attn_drop_rate,
                "proj_drop": self.proj_drop_rate,
                "norm_layer": keras.saving.serialize_keras_object(
                    self.norm_layer
                ),
                "use_bias": self.use_bias,
                "channel_axis": self.channel_axis,
                "data_format": self.data_format,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config["norm_layer"] = keras.saving.deserialize_keras_object(
            config["norm_layer"]
        )
        return cls(**config)


class Attention2d(keras.layers.Layer):
    """Implements 2D Multi-Head Attention.

    This layer performs multi-head self-attention on 2D spatial inputs.

    Args:
        filters: int. The output channel dimension.
        num_heads: int. The number of attention heads.
        bias: bool. If `True`, bias terms are used in the qkv and projection
            convolutions.
        attn_drop: float. The dropout rate for the attention weights.
        proj_drop: float. The dropout rate for the output projection.
        channel_axis: int. The axis representing the channels in the input
            tensor.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
    """

    def __init__(
        self,
        filters,
        num_heads=32,
        bias=True,
        attn_drop=0.0,
        proj_drop=0.0,
        channel_axis=None,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.num_heads = num_heads
        self.bias = bias
        self.attn_drop_rate = attn_drop
        self.proj_drop_rate = proj_drop
        self.channel_axis = channel_axis
        self.data_format = data_format
        self.conv_kernel_initializer = keras.initializers.VarianceScaling(
            scale=2.0, mode="fan_out", distribution="untruncated_normal"
        )
        self.bias_initializer = "zeros"
        self.attn_drop_layer = keras.layers.Dropout(
            attn_drop, dtype=self.dtype_policy
        )

    def build(self, input_shape):
        super().build(input_shape)
        dim = input_shape[self.channel_axis]
        self.head_dim = dim // self.num_heads
        self.qkv = keras.layers.Conv2D(
            dim * 3,
            kernel_size=1,
            use_bias=self.bias,
            data_format=self.data_format,
            name="qkv",
            dtype=self.dtype_policy,
            kernel_initializer=self.conv_kernel_initializer,
            bias_initializer=self.bias_initializer,
        )
        self.proj = keras.layers.Conv2D(
            self.filters,
            kernel_size=1,
            use_bias=self.bias,
            data_format=self.data_format,
            name="proj",
            dtype=self.dtype_policy,
            kernel_initializer=self.conv_kernel_initializer,
            bias_initializer=self.bias_initializer,
        )
        self.proj_drop_layer = keras.layers.Dropout(
            self.proj_drop_rate, dtype=self.dtype_policy
        )

    def call(self, x, attn_mask=None, training=False):
        if self.data_format == "channels_first":
            B, C, H, W = keras.ops.shape(x)
        else:
            B, H, W, C = keras.ops.shape(x)
        qkv = self.qkv(x)
        if self.data_format == "channels_last":
            qkv = keras.ops.transpose(qkv, (0, 3, 1, 2))
        q, k, v = keras.ops.unstack(
            keras.ops.reshape(
                qkv,
                (B, 3, self.num_heads, self.head_dim, H * W),
            ),
            axis=1,
        )
        q = keras.ops.transpose(q, (0, 1, 3, 2))
        k = keras.ops.transpose(k, (0, 1, 2, 3))
        v = keras.ops.transpose(v, (0, 1, 3, 2))
        attn = keras.ops.matmul(q, k) * (self.head_dim**-0.5)
        if attn_mask is not None:
            attn = attn + attn_mask
        attn = keras.ops.softmax(attn, axis=-1)
        attn = self.attn_drop_layer(attn, training=training)
        x = keras.ops.matmul(attn, v)
        x = keras.ops.transpose(x, (0, 1, 3, 2))
        if self.data_format == "channels_first":
            x = keras.ops.reshape(x, (B, -1, H, W))
        else:
            x = keras.ops.reshape(x, (B, H, W, -1))
        x = self.proj(x)
        x = self.proj_drop_layer(x, training=training)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "num_heads": self.num_heads,
                "bias": self.bias,
                "attn_drop": self.attn_drop_rate,
                "proj_drop": self.proj_drop_rate,
                "channel_axis": self.channel_axis,
                "data_format": self.data_format,
            }
        )
        return config


class MobileAttention(keras.layers.Layer):
    """MobileNetV5 attention block.

    This block combines attention with depthwise convolutions for efficiency.
    It can use either standard Multi-Head Attention or Multi-Query Attention.

    Args:
        filters: int. The number of output channels.
        stride: int. The stride for the block.
        dw_kernel_size: int. The kernel size for the depthwise convolution in
            Multi-Query Attention.
        dilation: int. The dilation rate for convolutions.
        pad_type: str. The padding type for convolutions.
        num_heads: int. The number of attention heads.
        key_dim: int. The dimension of the key.
        value_dim: int. The dimension of the value.
        use_multi_query: bool. If `True`, use `MultiQueryAttention2d`,
            otherwise use `Attention2d`.
        query_strides: tuple. The strides for the query downsampling.
        kv_stride: int. The stride for key/value downsampling.
        cpe_dw_kernel_size: int. The kernel size for the conditional position
            encoding depthwise convolution.
        noskip: bool. If `True`, the skip connection is disabled.
        norm_layer: str. The normalization layer to use (`"batch_norm"` or
            `"rms_norm"`).
        drop_path_rate: float. The stochastic depth rate.
        attn_drop: float. The dropout rate for the attention weights.
        proj_drop: float. The dropout rate for the output projection.
        layer_scale_init_value: float. The initial value for layer scale. If
            `None`, layer scale is not used.
        use_bias: bool. If `True`, bias terms are used in convolutions.
        use_cpe: bool. If `True`, a conditional position encoding is added.
        channel_axis: int. The axis representing the channels in the input
            tensor.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
    """

    def __init__(
        self,
        filters,
        stride=1,
        dw_kernel_size=3,
        dilation=1,
        pad_type="same",
        num_heads=8,
        key_dim=64,
        value_dim=64,
        use_multi_query=False,
        query_strides=(1, 1),
        kv_stride=1,
        cpe_dw_kernel_size=3,
        noskip=False,
        norm_layer="batch_norm",
        drop_path_rate=0.0,
        attn_drop=0.0,
        proj_drop=0.0,
        layer_scale_init_value=1e-5,
        use_bias=False,
        use_cpe=False,
        channel_axis=None,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.stride = stride
        self.dw_kernel_size = dw_kernel_size
        self.dilation = dilation
        self.pad_type = pad_type
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.value_dim = value_dim
        self.use_multi_query = use_multi_query
        self.query_strides = query_strides
        self.kv_stride = kv_stride
        self.cpe_dw_kernel_size = cpe_dw_kernel_size
        self.noskip = noskip
        self.norm_layer_name = norm_layer
        self.drop_path_rate = drop_path_rate
        self.attn_drop_rate = attn_drop
        self.proj_drop_rate = proj_drop
        self.layer_scale_init_value = layer_scale_init_value
        self.use_bias = use_bias
        self.use_cpe = use_cpe
        self.channel_axis = channel_axis
        self.data_format = data_format
        self.conv_kernel_initializer = keras.initializers.VarianceScaling(
            scale=2.0, mode="fan_out", distribution="untruncated_normal"
        )
        self.bias_initializer = "zeros"

    def build(self, input_shape):
        super().build(input_shape)
        in_chs = input_shape[self.channel_axis]
        self.has_skip = (
            self.stride == 1 and in_chs == self.filters
        ) and not self.noskip
        if self.use_cpe:
            self.conv_cpe_dw = keras.layers.DepthwiseConv2D(
                kernel_size=self.cpe_dw_kernel_size,
                strides=1,
                padding="same",
                dilation_rate=self.dilation,
                use_bias=True,
                data_format=self.data_format,
                name="conv_cpe_dw",
                depthwise_initializer=self.conv_kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
            )
        else:
            self.conv_cpe_dw = None
        if self.norm_layer_name == "batch_norm":
            self.norm = keras.layers.BatchNormalization(
                axis=self.channel_axis,
                name="norm",
                gamma_initializer="ones",
                beta_initializer="zeros",
                dtype=self.dtype_policy,
            )
        elif self.norm_layer_name == "rms_norm":
            self.norm = RmsNorm2d(
                in_chs,
                data_format=self.data_format,
                gamma_initializer="ones",
                channel_axis=self.channel_axis,
                name="norm",
                dtype=self.dtype_policy,
            )
        else:
            raise ValueError(f"Unsupported norm_layer: {self.norm_layer_name}")
        num_heads = self.num_heads
        if num_heads is None:
            assert in_chs % self.key_dim == 0
            num_heads = in_chs // self.key_dim
        attn_norm_layer = (
            RmsNorm2d
            if self.norm_layer_name == "rms_norm"
            else keras.layers.BatchNormalization
        )
        if self.use_multi_query:
            self.attn = MultiQueryAttention2d(
                filters=self.filters,
                num_heads=num_heads,
                key_dim=self.key_dim,
                value_dim=self.value_dim,
                query_strides=self.query_strides,
                kv_stride=self.kv_stride,
                dw_kernel_size=self.dw_kernel_size,
                dilation=self.dilation,
                padding=self.pad_type,
                attn_drop=self.attn_drop_rate,
                proj_drop=self.proj_drop_rate,
                norm_layer=attn_norm_layer,
                use_bias=self.use_bias,
                channel_axis=self.channel_axis,
                data_format=self.data_format,
                name="attn",
                dtype=self.dtype_policy,
            )
        else:
            self.attn = Attention2d(
                filters=self.filters,
                num_heads=num_heads,
                attn_drop=self.attn_drop_rate,
                proj_drop=self.proj_drop_rate,
                bias=self.use_bias,
                channel_axis=self.channel_axis,
                data_format=self.data_format,
                name="attn",
                dtype=self.dtype_policy,
            )
        if self.layer_scale_init_value is not None:
            self.layer_scale = LayerScale2d(
                self.filters,
                self.layer_scale_init_value,
                name="layer_scale",
                channel_axis=self.channel_axis,
                data_format=self.data_format,
                dtype=self.dtype_policy,
            )
        else:
            self.layer_scale = lambda x: x
        self.drop_path = (
            DropPath(self.drop_path_rate, dtype=self.dtype_policy)
            if self.drop_path_rate > 0.0
            else lambda x, training: x
        )

    def call(self, x, training=False):
        if self.conv_cpe_dw is not None:
            x = x + self.conv_cpe_dw(x)
        shortcut = x
        x_normed = self.norm(x, training=training)
        x_attn = self.attn(x_normed, training=training)
        x_scaled = self.layer_scale(x_attn)
        if self.has_skip:
            return self.drop_path(x_scaled, training=training) + shortcut
        else:
            return x_scaled

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "stride": self.stride,
                "dw_kernel_size": self.dw_kernel_size,
                "dilation": self.dilation,
                "pad_type": self.pad_type,
                "num_heads": self.num_heads,
                "key_dim": self.key_dim,
                "value_dim": self.value_dim,
                "use_multi_query": self.use_multi_query,
                "query_strides": self.query_strides,
                "kv_stride": self.kv_stride,
                "cpe_dw_kernel_size": self.cpe_dw_kernel_size,
                "noskip": self.noskip,
                "norm_layer": self.norm_layer_name,
                "drop_path_rate": self.drop_path_rate,
                "attn_drop": self.attn_drop_rate,
                "proj_drop": self.proj_drop_rate,
                "layer_scale_init_value": self.layer_scale_init_value,
                "use_bias": self.use_bias,
                "use_cpe": self.use_cpe,
                "channel_axis": self.channel_axis,
                "data_format": self.data_format,
            }
        )
        return config



"""DeiT model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "deit_base_distilled_patch16_384_imagenet": {
        "metadata": {
            "description": (
                "DeiT-B16 model pre-trained on the ImageNet 1k dataset with "
                "image resolution of 384x384 "
            ),
            "params": 86_092_032,
            "path": "deit",
        },
        "kaggle_handle": "kaggle://keras/deit/keras/deit_base_distilled_patch16_384_imagenet/1",
    },
    "deit_base_distilled_patch16_224_imagenet": {
        "metadata": {
            "description": (
                "DeiT-B16 model pre-trained on the ImageNet 1k dataset with "
                "image resolution of 224x224 "
            ),
            "params": 85_800_192,
            "path": "deit",
        },
        "kaggle_handle": "kaggle://keras/deit/keras/deit_base_distilled_patch16_224_imagenet/1",
    },
    "deit_tiny_distilled_patch16_224_imagenet": {
        "metadata": {
            "description": (
                "DeiT-T16 model pre-trained on the ImageNet 1k dataset with "
                "image resolution of 224x224 "
            ),
            "params": 5_524_800,
            "path": "deit",
        },
        "kaggle_handle": "kaggle://keras/deit/keras/deit_tiny_distilled_patch16_224_imagenet/1",
    },
    "deit_small_distilled_patch16_224_imagenet": {
        "metadata": {
            "description": (
                "DeiT-S16 model pre-trained on the ImageNet 1k dataset with "
                "image resolution of 224x224 "
            ),
            "params": 21_666_432,
            "path": "deit",
        },
        "kaggle_handle": "kaggle://keras/deit/keras/deit_small_distilled_patch16_224_imagenet/1",
    },
}



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.depth_anything.depth_anything_backbone import (
    DepthAnythingBackbone,
)
from keras_hub.src.models.depth_anything.depth_anything_depth_estimator_preprocessor import (  # noqa: E501
    DepthAnythingDepthEstimatorPreprocessor,
)
from keras_hub.src.models.depth_anything.depth_anything_loss import (
    DepthAnythingLoss,
)
from keras_hub.src.models.depth_estimator import DepthEstimator


@keras_hub_export("keras_hub.models.DepthAnythingDepthEstimator")
class DepthAnythingDepthEstimator(DepthEstimator):
    backbone_cls = DepthAnythingBackbone
    preprocessor_cls = DepthAnythingDepthEstimatorPreprocessor

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        *,
        metrics="auto",
        **kwargs,
    ):
        """Configures the `DepthEstimator` task for training.

        The `DepthEstimator` task extends the default compilation signature of
        `keras.Model.compile` with defaults for `optimizer`, `loss`, and
        `metrics`. To override these defaults, pass any value
        to these arguments during compilation.

        Args:
            optimizer: `"auto"`, an optimizer name, or a `keras.Optimizer`
                instance. Defaults to `"auto"`, which uses the default optimizer
                for the given model and task. See `keras.Model.compile` and
                `keras.optimizers` for more info on possible `optimizer` values.
            loss: `"auto"`, a loss name, or a `keras.losses.Loss` instance.
                Defaults to `"auto"`, where a `DepthAnythingLoss` loss will be
                applied for the depth estimation task. See
                `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            metrics: `"auto"`, or a dict of metrics to be evaluated by
                the model during training and testing. Defaults to `"auto"`,
                where a `keras.metrics.RootMeanSquaredError` will be applied to
                track the accuracy of the model during training. See
                `keras.Model.compile` and `keras.metrics` for more info on
                possible `metrics` values.
            **kwargs: See `keras.Model.compile` for a full list of arguments
                supported by the compile method.
        """
        if optimizer == "auto":
            optimizer = keras.optimizers.AdamW(5e-5)
        if loss == "auto":
            loss = {
                "depths": DepthAnythingLoss(
                    min_depth=self.min_depth, max_depth=self.max_depth
                )
            }
        if metrics == "auto":
            metrics = {"depths": keras.metrics.RootMeanSquaredError()}
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            **kwargs,
        )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_segmenter_preprocessor import (
    ImageSegmenterPreprocessor,
)
from keras_hub.src.models.segformer.segformer_backbone import SegFormerBackbone
from keras_hub.src.models.segformer.segformer_image_converter import (
    SegFormerImageConverter,
)


@keras_hub_export("keras_hub.models.SegFormerImageSegmenterPreprocessor")
class SegFormerImageSegmenterPreprocessor(ImageSegmenterPreprocessor):
    backbone_cls = SegFormerBackbone
    image_converter_cls = SegFormerImageConverter



from keras_hub.src.models.gpt2.gpt2_backbone import GPT2Backbone
from keras_hub.src.models.gpt2.gpt2_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, GPT2Backbone)



"""Xception preset configurations."""

backbone_presets = {
    "xception_41_imagenet": {
        "metadata": {
            "description": (
                "41-layer Xception model pre-trained on ImageNet 1k."
            ),
            "params": 20861480,
            "path": "xception",
        },
        "kaggle_handle": "kaggle://keras/xception/keras/xception_41_imagenet/2",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.efficientnet.efficientnet_backbone import (
    EfficientNetBackbone,
)
from keras_hub.src.models.efficientnet.efficientnet_image_converter import (
    EfficientNetImageConverter,
)
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)


@keras_hub_export("keras_hub.models.EfficientNetImageClassifierPreprocessor")
class EfficientNetImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = EfficientNetBackbone
    image_converter_cls = EfficientNetImageConverter



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.bert.bert_tokenizer import BertTokenizer
from keras_hub.src.models.roformer_v2.roformer_v2_backbone import (
    RoformerV2Backbone,
)


@keras_hub_export(
    [
        "keras_hub.tokenizers.RoformerV2Tokenizer",
        "keras_hub.models.RoformerV2Tokenizer",
    ]
)
class RoformerV2Tokenizer(BertTokenizer):
    """A RoformerV2 tokenizer using WordPiece subword segmentation.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_hub.tokenizers.WordPieceTokenizer`. Unlike the
    underlying tokenizer, it will check for special tokens needed by RoformerV2
    models and provides a `from_preset()` method to automatically download
    a matching vocabulary for a RoformerV2 preset.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.

    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Args:
        vocabulary: A list of strings or a string filename path. If
            passing a list, each element of the list should be a single word
            piece token string. If passing a filename, the file should be a
            plain text file containing a single word piece token per line.
        lowercase: If `True`, the input text will be first lowered before
            tokenization.
        special_tokens_in_strings: bool. A bool to indicate if the tokenizer
            should expect special tokens in input strings that should be
            tokenized and mapped correctly to their ids. Defaults to False.

    Examples:
    ```python
    # Unbatched input.
    tokenizer = keras_hub.models.RoformerV2Tokenizer.from_preset(
        "roformer_v2_base_zh",
    )
    tokenizer("The quick brown fox jumped.")

    # Batched input.
    tokenizer(["The quick brown fox jumped.", "The fox slept."])

    # Detokenization.
    tokenizer.detokenize(tokenizer("The quick brown fox jumped."))

    # Custom vocabulary.
    vocab = ["[UNK]", "[CLS]", "[SEP]", "[PAD]", "[MASK]"]
    vocab += ["The", "quick", "brown", "fox", "jumped", "."]
    tokenizer = keras_hub.models.RoformerV2Tokenizer(vocabulary=vocab)
    tokenizer("The quick brown fox jumped.")
    ```
    """

    backbone_cls = RoformerV2Backbone



import keras
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.modeling.reversible_embedding import (
    ReversibleEmbedding,
)
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.qwen.qwen_layernorm import QwenLayerNorm
from keras_hub.src.models.qwen_moe.qwen_moe_decoder import (
    QwenMoeTransformerDecoder,
)


def _qwen_moe_kernel_initializer(stddev=0.02):
    return keras.initializers.RandomNormal(stddev=stddev)


@keras_hub_export(
    "keras_hub.models.QwenMoeBackbone",
)
class QwenMoeBackbone(Backbone):
    """Qwen MoE core network with hyperparameters.

    This backbone implements the base Transformer network for the Qwen MoE
    model. It includes embedding lookups and transformer layers with a Mixture
    of Experts (MoE) architecture, where each layer uses a sparse set of experts
    for efficient computation. This backbone outputs the final hidden states for
    each token, not generative predictions over the vocabulary space. For higher
    -level object for text generation, see `keras_hub.models.QwenMoeCausalLM`.

    The default constructor gives a fully customizable, randomly initialized
    Qwen MoE model with any number of layers, heads, and embedding dimensions.
    To load preset architectures and weights, use the `from_preset` constructor.

    Args:
        vocabulary_size: int. The size of the token vocabulary.
        num_layers: int. The number of transformer layers.
        num_query_heads: int. The number of heads for the query projections in
            the attention layer.
        num_key_value_heads: int. The number of heads for the key and value
            projections in the attention layer.
        hidden_dim: int. The size of the transformer hidden state at the end of
            each transformer layer.
        intermediate_dim: int. The output dimension of the first Dense layer in
            the feedforward network for each transformer.
        moe_intermediate_dim: int. The intermediate dimension for each expert
            in the MoE feedforward network.
        shared_expert_intermediate_dim: int. The intermediate dimension for the
            shared expert in the MoE feedforward network.
        num_experts: int. The number of experts in each MoE layer.
        top_k: int. The number of top experts to select for each token in the
            MoE layer.
        head_dim: int. The size of each attention head.
        layer_norm_epsilon: float. The epsilon value used for every layer norm
            in the transformer model.
        dropout: float. Dropout probability for the transformer encoder.
        use_sliding_window_attention: bool. Whether to use sliding local window
            attention. Defaults to False.
        sliding_window_size: int. Size of the sliding local window. Defaults to
            4096.
        max_sequence_length: int. The maximum sequence length supported by the
            model. Defaults to 4096.
        dtype: str or `keras.mixed_precision.DTypePolicy`. The dtype to use for
            the model's computations and weights. Note that some computations,
            such as softmax and layer normalization, will always be done at
            float32 precision regardless of dtype.

    Example:
    ```python
    input_data = {
        "token_ids": np.ones(shape=(1, 12), dtype="int32"),
        "padding_mask": np.array([[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0]]),
    }

    # Pretrained Qwen MoE decoder.
    model = keras_hub.models.QwenMoeBackbone.from_preset("qwen_moe_a2_7b")
    model(input_data)

    # Randomly initialized Qwen MoE decoder with custom config.
    model = keras_hub.models.QwenMoeBackbone(
        vocabulary_size=151936,
        num_layers=28,
        num_query_heads=16,
        num_key_value_heads=8,
        hidden_dim=2048,
        intermediate_dim=4096,
        moe_intermediate_dim=128,
        shared_expert_intermediate_dim=4096,
        num_experts=60,
        top_k=4,
        head_dim=128,
        max_sequence_length=4096,
    )
    model(input_data)
    """

    def __init__(
        self,
        vocabulary_size,
        num_layers,
        num_query_heads,
        num_key_value_heads,
        hidden_dim,
        intermediate_dim,
        moe_intermediate_dim,
        shared_expert_intermediate_dim,
        num_experts,
        top_k=4,
        norm_top_k_prob=False,
        decoder_sparse_step=1,
        rope_max_wavelength=10000,
        rope_scaling_factor=1.0,
        layer_norm_epsilon=1e-6,
        dropout=0,
        dtype=None,
        tie_word_embeddings=False,
        use_sliding_window_attention=False,
        sliding_window_size=32768,
        output_router_logits=False,
        router_aux_loss_coefficient=0.001,
        mlp_only_layers=[],
        training=None,
        **kwargs,
    ):
        # === Layers ===
        self.token_embedding = ReversibleEmbedding(
            input_dim=vocabulary_size,
            output_dim=hidden_dim,
            tie_weights=tie_word_embeddings,
            embeddings_initializer=_qwen_moe_kernel_initializer(stddev=0.01),
            dtype=dtype,
            name="token_embedding",
        )
        self.transformer_layers = []
        for i in range(num_layers):
            layer = QwenMoeTransformerDecoder(
                intermediate_dim=intermediate_dim,
                num_query_heads=num_query_heads,
                num_key_value_heads=num_key_value_heads,
                moe_intermediate_dim=moe_intermediate_dim,
                shared_expert_intermediate_dim=shared_expert_intermediate_dim,
                num_experts=num_experts,
                top_k=top_k,
                norm_top_k_prob=norm_top_k_prob,
                decoder_sparse_step=decoder_sparse_step,
                rope_max_wavelength=rope_max_wavelength,
                rope_scaling_factor=rope_scaling_factor,
                layer_norm_epsilon=layer_norm_epsilon,
                activation=ops.silu,
                kernel_initializer=_qwen_moe_kernel_initializer(stddev=0.02),
                dropout=dropout,
                dtype=dtype,
                use_sliding_window_attention=use_sliding_window_attention,
                sliding_window_size=sliding_window_size,
                output_router_logits=output_router_logits,
                router_aux_loss_coefficient=router_aux_loss_coefficient,
                mlp_only_layers=mlp_only_layers,
                name=f"transformer_layer_{i}",
            )
            self.transformer_layers.append(layer)
        self.layer_norm = QwenLayerNorm(
            epsilon=layer_norm_epsilon,
            dtype=dtype,
            name="sequence_output_layernorm",
        )

        # === Functional Model ===
        token_id_input = keras.Input(
            shape=(None,), dtype="int32", name="token_ids"
        )
        padding_mask_input = keras.Input(
            shape=(None,), dtype="int32", name="padding_mask"
        )
        x = self.token_embedding(token_id_input)
        for transformer_layer in self.transformer_layers:
            x = transformer_layer(
                x, decoder_padding_mask=padding_mask_input, training=training
            )
        sequence_output = self.layer_norm(x)
        super().__init__(
            inputs={
                "token_ids": token_id_input,
                "padding_mask": padding_mask_input,
            },
            outputs=sequence_output,
            dtype=dtype,
            **kwargs,
        )

        # === Config ===
        self.vocabulary_size = vocabulary_size
        self.num_layers = num_layers
        self.num_query_heads = num_query_heads
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.moe_intermediate_dim = moe_intermediate_dim
        self.shared_expert_intermediate_dim = shared_expert_intermediate_dim
        self.rope_max_wavelength = rope_max_wavelength
        self.num_key_value_heads = num_key_value_heads
        self.rope_scaling_factor = rope_scaling_factor
        self.layer_norm_epsilon = layer_norm_epsilon
        self.dropout = dropout
        self.tie_word_embeddings = tie_word_embeddings
        self.use_sliding_window_attention = use_sliding_window_attention
        self.sliding_window_size = sliding_window_size
        self.num_experts = num_experts
        self.top_k = top_k
        self.norm_top_k_prob = norm_top_k_prob
        self.decoder_sparse_step = decoder_sparse_step
        self.mlp_only_layers = mlp_only_layers
        self.router_aux_loss_coefficient = router_aux_loss_coefficient
        self.output_router_logits = output_router_logits

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary_size": self.vocabulary_size,
                "num_layers": self.num_layers,
                "num_query_heads": self.num_query_heads,
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "moe_intermediate_dim": self.moe_intermediate_dim,
                "shared_expert_intermediate_dim": (
                    self.shared_expert_intermediate_dim
                ),
                "rope_max_wavelength": self.rope_max_wavelength,
                "num_key_value_heads": self.num_key_value_heads,
                "rope_scaling_factor": self.rope_scaling_factor,
                "layer_norm_epsilon": self.layer_norm_epsilon,
                "dropout": self.dropout,
                "tie_word_embeddings": self.tie_word_embeddings,
                "use_sliding_window_attention": (
                    self.use_sliding_window_attention
                ),
                "sliding_window_size": self.sliding_window_size,
                "num_experts": self.num_experts,
                "top_k": self.top_k,
                "norm_top_k_prob": self.norm_top_k_prob,
                "decoder_sparse_step": self.decoder_sparse_step,
                "mlp_only_layers": self.mlp_only_layers,
                "output_router_logits": self.output_router_logits,
            }
        )
        return config

    @staticmethod
    def get_layout_map(
        device_mesh,
        model_parallel_dim_name="model",
        data_parallel_dim_name="batch",
    ):
        """Get a `keras.distribution.LayoutMap` for model parallel distribution.

        The returned `LayoutMap` contains the sharding spec for the QwenMoe
        backbone weights, so that you can use it to distribute weights across
        the accelerators.

        Example:
        ```
        # Feel free to change the mesh shape to balance data and model
        # parallelism
        mesh = keras.distribution.DeviceMesh(
            shape=(1, 8),
            axis_names=('batch', 'model'),
            devices=keras.distribution.list_devices(),
        )
        layout_map = QwenMoeBackbone.get_layout_map(
            mesh,
            model_parallel_dim_name="model",
        )

        distribution = keras.distribution.ModelParallel(
            layout_map=layout_map,
            batch_dim_name='batch',
        )

        with distribution.scope():
           qwen_moe_model = keras_hub.models.QwenMoeBackbone.from_preset()
        ```

        To see how the layout map was applied, load the model then run
        (for one decoder block):
        ```
        embedding_layer = qwen_moe_model.backbone.get_layer("token_embedding")
        decoder_block_1 = qwen_moe_model.backbone.get_layer(
            'transformer_layer_0'
        )
        for variable in embedding_layer.weights + decoder_block_1.weights:
            print(
                f'{variable.path:<58}  {str(variable.shape):<16}  '
                f'{str(variable.value.sharding.spec)}'
            )
        ```

        Args:
            device_mesh: The `keras.distribution.DeviceMesh` instance for
                distribution.
            model_parallel_dim_name: The axis name of the device mesh, where
                the weights should be partition on.
            data_parallel_dim_name: The axis name of the device mesh, where
                the data should be partition on.
        Return:
            `keras.distribution.LayoutMap` that contains the sharding spec
            for all the model weights.
        """
        # The weight path and shape of the Llama backbone is like below
        # token_embedding/embeddings                              (128256, 2048)
        # repeat block for decoder
        # transformer_layer_0/self_attention/query/kernel         (2048, 32, 64)
        # transformer_layer_0/self_attention/key/kernel           (2048, 8, 64)
        # transformer_layer_0/self_attention/value/kernel         (2048, 8, 64)
        # transformer_layer_0/self_attention/attention_output/kernel
        #                                                         (32, 64, 2048)
        # transformer_layer_0/self_attention_layernorm/scale      (2048,)
        # transformer_layer_0/feedforward_intermediate_dense/kernel
        #                                                         (2048, 8192)
        # transformer_layer_0/feedforward_gate_dense/kernel       (2048, 8192)
        # transformer_layer_0/feedforward_output_dense/kerne      (8192, 2048)
        # transformer_layer_0/feedforward_layernorm/scale         (2048,)

        if not isinstance(device_mesh, keras.distribution.DeviceMesh):
            raise ValueError(
                "Invalid device_mesh type. Expected "
                f"`keras.distribution.Device`, got {type(device_mesh)}"
            )
        if model_parallel_dim_name not in device_mesh.axis_names:
            raise ValueError(
                f"{model_parallel_dim_name} is not found in the "
                f"device_mesh.axis_names. {device_mesh.axis_name=}"
            )
        if data_parallel_dim_name not in device_mesh.axis_names:
            raise ValueError(
                f"{data_parallel_dim_name} is not found in the "
                f"device_mesh.axis_names. {device_mesh.axis_name=}"
            )
        # Note that it is possible to further config the mesh to be 3D, eg
        # (data, seq, model). We leave it as 2D for now for simplicity.
        data_dim = data_parallel_dim_name
        model_dim = model_parallel_dim_name
        # The sharding config is based on the Gemma team training config.
        # See https://arxiv.org/abs/2403.08295
        layout_map = keras.distribution.LayoutMap(device_mesh)
        layout_map["token_embedding/embeddings"] = (model_dim, data_dim)
        layout_map[
            "transformer_layer.*self_attention.*(query|key|value).kernel"
        ] = (
            model_dim,
            data_dim,
            None,
        )
        layout_map["transformer_layer.*attention_output.kernel"] = (
            model_dim,
            None,
            data_dim,
        )
        layout_map[
            "transformer_layer.*feedforward_intermediate_dense.kernel"
        ] = (
            data_dim,
            model_dim,
        )
        layout_map["transformer_layer.*feedforward_gate_dense.kernel"] = (
            data_dim,
            model_dim,
        )
        layout_map["transformer_layer.*feedforward_output_dense.kernel"] = (
            model_dim,
            data_dim,
        )

        return layout_map



import keras
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm import CausalLM
from keras_hub.src.models.smollm3.smollm3_backbone import SmolLM3Backbone
from keras_hub.src.models.smollm3.smollm3_causal_lm_preprocessor import (
    SmolLM3CausalLMPreprocessor,
)
from keras_hub.src.utils.tensor_utils import any_equal


@keras_hub_export(
    [
        "keras_hub.models.SmolLM3CausalLM",
        "keras_hub.models.SmolLMCausalLM",
    ]
)
class SmolLM3CausalLM(CausalLM):
    backbone_cls = SmolLM3Backbone
    preprocessor_cls = SmolLM3CausalLMPreprocessor

    def __init__(self, backbone, preprocessor=None, **kwargs):
        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor

        # === Functional Model ===
        # This must be "backbone.input" i.e. the full input structure,
        # rather than "backbone.inputs" which is the flattened list of inputs.
        inputs = backbone.input
        hidden_states = backbone(inputs)
        outputs = backbone.token_embedding(hidden_states, reverse=True)
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

    def call_with_cache(
        self,
        token_ids,
        cache,
        cache_update_index,
    ):
        """Forward pass of `SmolLM3CausalLM` with cache.

        `call_with_cache` adds an additional forward pass for the model for
        autoregressive inference. Unlike calling the model directly, this method
        allows caching previous key/value Tensors in multi-head attention layer,
        and avoids recomputing the outputs of seen tokens.

        Args:
            token_ids: a dense int Tensor with shape `(batch_size, seq_len)`.
                For prefill, `seq_len` is the prompt length. For generation,
                `seq_len` is typically 1.
            cache: a dense float Tensor, the cache of key and value.
                Shape: (batch_size, num_layers, 2, max_seq_len,
                num_key_value_heads, head_dim)
            cache_update_index: int, or int Tensor. The index of current
                inputs in the whole sequence.
            training: Boolean, whether the call is during training or inference.
            attention_mask: Optional attention mask.

        Returns:
            A (logits, hidden_states, cache) tuple. Where `logits` is the
            language model logits for the input token_ids, `hidden_states` is
            the final hidden representation of the input tokens, and `cache` is
            the decoding cache.
        """
        x = self.backbone.token_embedding(token_ids)

        # Each decoder layer has a cache; we update them separately.
        updated_cache = []

        for i in range(self.backbone.num_layers):
            current_cache = cache[:, i, ...]
            x, next_cache = self.backbone.transformer_layers[i](
                x,
                self_attention_cache=current_cache,
                self_attention_cache_update_index=cache_update_index,
            )
            updated_cache.append(next_cache)
        cache = ops.stack(updated_cache, axis=1)
        hidden_states = x = self.backbone.norm(x)
        logits = self.backbone.token_embedding(x, reverse=True)
        return logits, hidden_states, cache

    def _build_cache(self, token_ids):
        """Build an empty cache for use with `call_with_cache()`."""
        batch_size = ops.shape(token_ids)[0]
        max_length = ops.shape(token_ids)[1]
        num_layers = self.backbone.num_layers
        num_key_value_heads = self.backbone.num_key_value_heads
        head_dim = self.backbone.hidden_dim // self.backbone.num_attention_heads
        shape = [
            batch_size,
            num_layers,
            2,
            max_length,
            num_key_value_heads,
            head_dim,
        ]
        cache = ops.zeros(shape, dtype=self.compute_dtype)
        index = ops.convert_to_tensor(0, dtype="int32")
        # Seed the cache.
        _, hidden_states, cache = self.call_with_cache(token_ids, cache, index)
        return hidden_states, cache

    def generate_step(
        self,
        inputs,
        stop_token_ids=None,
    ):
        """A compilable generation function for a single batch of inputs.

        This function represents the inner, XLA-compilable, generation function
        for a single batch of inputs. Inputs should have the same structure as
        model inputs, a dictionary with keys `"token_ids"` and `"padding_mask"`.

        Args:
            inputs: A dictionary with two keys `"token_ids"` and
                `"padding_mask"` and batched tensor values.
            stop_token_ids: Tuple of id's of the end token to stop on. If all
                sequences have produced a new stop token, generation
                will stop.
        """
        token_ids, padding_mask = inputs["token_ids"], inputs["padding_mask"]

        hidden_states, cache = self._build_cache(token_ids)
        # Compute the lengths of all user inputted tokens ids.
        row_lengths = ops.sum(ops.cast(padding_mask, "int32"), axis=-1)
        # Start at the first index that has no user inputted id.
        index = ops.min(row_lengths)

        def next(prompt, cache, index):
            # The cache index is the index of our previous token.
            cache_update_index = index - 1
            batch_size = ops.shape(prompt)[0]
            prompt = ops.slice(prompt, [0, cache_update_index], [batch_size, 1])

            logits, hidden_states, cache = self.call_with_cache(
                prompt,
                cache,
                cache_update_index,
            )
            return (
                ops.squeeze(logits, axis=1),
                ops.squeeze(hidden_states, axis=1),
                cache,
            )

        token_ids = self.sampler(
            next=next,
            prompt=token_ids,
            cache=cache,
            index=index,
            mask=padding_mask,
            stop_token_ids=stop_token_ids,
            hidden_states=hidden_states,
            model=self,
        )

        # Compute an output padding mask with the token ids we updated.
        if stop_token_ids is not None:
            # Build a mask of stop token locations not in the original
            # prompt (not in locations where `padding_mask` is True).
            end_locations = any_equal(
                token_ids, stop_token_ids, ops.logical_not(padding_mask)
            )
            end_locations = ops.cast(end_locations, "int32")
            # Use cumsum to get ones in all locations after end_locations.
            cumsum = ops.cast(ops.cumsum(end_locations, axis=-1), "int32")
            overflow = cumsum - end_locations
            # Our padding mask is the inverse of these overflow locations.
            padding_mask = ops.logical_not(ops.cast(overflow, "bool"))
        else:
            # Without early stopping, all locations will have been updated.
            padding_mask = ops.ones_like(token_ids, dtype="bool")
        return {
            "token_ids": token_ids,
            "padding_mask": padding_mask,
        }

    def score(
        self,
        token_ids,
        padding_mask=None,
        scoring_mode="logits",
        layer_intercept_fn=None,
        target_ids=None,
    ):
        """Score a generation represented by the provided token ids.

        Args:
            token_ids: A <int>[batch_size, num_tokens] tensor containing tokens
                to score. Typically, this tensor captures the output from a call
                to `SmolLM3CausalLM.generate()`, i.e., tokens for both the input
                text and the model-generated text.
            padding_mask: A <bool>[batch_size, num_tokens] tensor indicating the
                tokens that should be preserved during generation. This is an
                artifact required by the `SmolLM3Backbone` and isn't influential
                on the computation of this function. If omitted, this function
                uses `keras.ops.ones()` to create a tensor of the appropriate
                shape.
            scoring_mode: The type of scores to return, either "logits" or
                "loss", both will be per input token.
            layer_intercept_fn: An optional function for augmenting activations
                with additional computation, for example, as part of
                interpretability research. This function will be passed the
                activations as its first parameter and a numeric index
                associated with that backbone layer. _This index _is not_ an
                index into `self.backbone.layers`_. The index -1 accompanies the
                embeddings returned by calling `self.backbone.token_embedding()`
                on `token_ids` in the forward direction. All subsequent indexes
                will be 0-based indices for the activations returned by each of
                the Transformers layers in the backbone. This function must
                return a <float>[batch_size, num_tokens, hidden_dims] tensor
                that can be passed as an input to the next layer in the model.
            target_ids: An <bool>[batch_size, num_tokens] tensor containing the
                predicted tokens against which the loss should be computed. If a
                span of tokens is provided (sequential truthy values along
                axis=1 in the tensor), the loss will be computed as the
                aggregate across those tokens.

        Raises:
            ValueError: If an unsupported scoring_mode is provided, or if the
                target_ids are not provided when using ScoringMode.LOSS.

        Returns:
            The per-token scores as a tensor of size
            <float>[batch_size, num_tokens, vocab_size] in "logits" mode, or
            <float>[batch_size, num_tokens] in "loss" mode.

        Example:

        Compute gradients between embeddings and loss scores with TensorFlow:
        ```python
        smol_lm = keras_hub.models.SmolLM3CausalLM.from_preset("...")
        generations = smol_lm.generate(
            ["This is a", "Where are you"],
            max_length=30
        )
        preprocessed = smol_lm.preprocessor.generate_preprocess(generations)
        generation_ids = preprocessed["token_ids"]
        padding_mask = preprocessed["padding_mask"]
        target_ids = keras.ops.roll(generation_ids, shift=-1, axis=1)

        embeddings = None
        with tf.GradientTape(watch_accessed_variables=True) as tape:
            def layer_intercept_fn(x, i):
                if i == -1:
                    nonlocal embeddings, tape
                    embeddings = x
                    tape.watch(embeddings)
                return x

            losses = smol_lm.score(
                token_ids=generation_ids,
                padding_mask=padding_mask,
                scoring_mode="loss",
                layer_intercept_fn=layer_intercept_fn,
                target_ids=target_ids,
            )

        grads = tape.gradient(losses, embeddings)
        ```
        """
        if scoring_mode not in ("logits", "loss"):
            raise ValueError(
                "Unsupported scoring_mode. Must be one of 'logits' or 'loss'."
            )

        if scoring_mode == "loss" and target_ids is None:
            raise ValueError(
                "Cannot compute loss without targets. Please provide target "
                "token ids via the target_ids parameter."
            )

        batch_shape = ops.shape(token_ids)[:2]
        assert len(batch_shape) == 2

        if padding_mask is None:
            padding_mask = ops.ones(shape=batch_shape)

        if layer_intercept_fn is None:

            def default_layer_intercept_fn(x, unused_i):
                return x

            layer_intercept_fn = default_layer_intercept_fn

        token_embeddings = self.backbone.token_embedding(token_ids)
        x = layer_intercept_fn(token_embeddings, -1)

        for i, transformer_layer in enumerate(self.backbone.transformer_layers):
            x = transformer_layer(x, decoder_padding_mask=padding_mask)
            x = layer_intercept_fn(x, i)

        x = self.backbone.norm(x)
        logits = self.backbone.token_embedding(x, reverse=True)

        if scoring_mode == "logits":
            return logits

        per_token_loss_fn = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction="none"
        )
        per_token_loss = per_token_loss_fn(target_ids, logits)
        return per_token_loss



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.deit.deit_layers import DeiTEmbeddings
from keras_hub.src.models.deit.deit_layers import DeiTEncoder
from keras_hub.src.utils.keras_utils import standardize_data_format


@keras_hub_export("keras_hub.models.DeiTBackbone")
class DeiTBackbone(Backbone):
    """DeiT backbone.

    This backbone implements the Data-efficient Image Transformer (DeiT)
    architecture as described in [Training data-efficient image
    transformers & distillation through attention]
    (https://arxiv.org/abs/2012.12877).

    Args:
        image_shape: A tuple or list of 3 integers representing the shape of the
            input image `(height, width, channels)`.
        patch_size: tuple or int. The size of each image patch. If an int is
            provided, it will be used for both height and width. The input image
            will be split into patches of shape `(patch_size_h, patch_size_w)`.
        num_layers: int. The number of transformer encoder layers.
        num_heads: int. The number of attention heads in each Transformer
            encoder layer.
        hidden_dim: int. The dimensionality of the hidden representations.
        intermediate_dim: int. The dimensionality of the intermediate MLP layer
            in each Transformer encoder layer.
        dropout_rate: float. The dropout rate for the Transformer encoder
            layers.
        attention_dropout: float. The dropout rate for the attention mechanism
            in each Transformer encoder layer.
        layer_norm_epsilon: float. Value used for numerical stability in layer
            normalization.
        use_mha_bias: bool. Whether to use bias in the multi-head attention
            layers.
        data_format: str. `"channels_last"` or `"channels_first"`, specifying
            the data format for the input image. If `None`, defaults to
            `"channels_last"`.
        dtype: The dtype of the layer weights. Defaults to None.
        **kwargs: Additional keyword arguments to be passed to the parent
            `Backbone` class.
    """

    def __init__(
        self,
        image_shape,
        patch_size,
        num_layers,
        num_heads,
        hidden_dim,
        intermediate_dim,
        dropout_rate=0.0,
        attention_dropout=0.0,
        layer_norm_epsilon=1e-6,
        use_mha_bias=True,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        # === Laters ===
        data_format = standardize_data_format(data_format)
        if isinstance(patch_size, int):
            patch_size = (patch_size, patch_size)
        h_axis, w_axis, channels_axis = (
            (-3, -2, -1) if data_format == "channels_last" else (-2, -1, -3)
        )
        # Check that the input image is well specified.
        if image_shape[h_axis] is None or image_shape[w_axis] is None:
            raise ValueError(
                f"Image shape must have defined height and width. Found `None` "
                f"at index {h_axis} (height) or {w_axis} (width). "
                f"Image shape: {image_shape}"
            )
        # Check that image dimensions be divisible by patch size
        if image_shape[h_axis] % patch_size[0] != 0:
            raise ValueError(
                f"Input height {image_shape[h_axis]} should be divisible by "
                f"patch size {patch_size}."
            )
        if image_shape[w_axis] % patch_size[1] != 0:
            raise ValueError(
                f"Input height {image_shape[w_axis]} should be divisible by "
                f"patch size {patch_size}."
            )

        num_channels = image_shape[channels_axis]

        # === Functional Model ===
        inputs = keras.layers.Input(shape=image_shape)

        x = DeiTEmbeddings(
            image_size=(image_shape[h_axis], image_shape[w_axis]),
            patch_size=patch_size,
            hidden_dim=hidden_dim,
            num_channels=num_channels,
            data_format=data_format,
            dropout_rate=dropout_rate,
            dtype=dtype,
            name="deit_patching_and_embedding",
        )(inputs)

        output, _, _ = DeiTEncoder(
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            intermediate_dim=intermediate_dim,
            use_mha_bias=use_mha_bias,
            dropout_rate=dropout_rate,
            attention_dropout=attention_dropout,
            layer_norm_epsilon=layer_norm_epsilon,
            dtype=dtype,
            name="deit_encoder",
        )(x)

        super().__init__(
            inputs=inputs,
            outputs=output,
            dtype=dtype,
            **kwargs,
        )

        # === Config ===
        self.image_shape = image_shape
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.dropout_rate = dropout_rate
        self.attention_dropout = attention_dropout
        self.layer_norm_epsilon = layer_norm_epsilon
        self.use_mha_bias = use_mha_bias
        self.data_format = data_format

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "image_shape": self.image_shape,
                "patch_size": self.patch_size,
                "num_layers": self.num_layers,
                "num_heads": self.num_heads,
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "dropout_rate": self.dropout_rate,
                "attention_dropout": self.attention_dropout,
                "layer_norm_epsilon": self.layer_norm_epsilon,
                "use_mha_bias": self.use_mha_bias,
            }
        )
        return config
